from typing import Tuple, Optional, Union
import jax
import jax.numpy as jnp
import chex
from flax import struct
from ..strategy import Strategy


@struct.dataclass
class EvoState:
    p_sigma: chex.Array
    mean: chex.Array
    sigma: float
    P: chex.Array
    t_gap: chex.Array
    s_rank_rate: float
    fitness_archive: chex.Array
    weights: chex.Array
    best_member: chex.Array
    best_fitness: float = jnp.finfo(jnp.float32).max
    gen_counter: int = 0


@struct.dataclass
class EvoParams:
    c_cov: float
    c_c: float
    c_sigma: float
    mu_eff: float
    c_m: float = 1.0
    sigma_init: float = 1.0
    sigma_limit: float = 0.001
    t_uncorr: int = 20
    q_star: float = 0.325
    c_s: float = 0.3
    d_sigma: float = 1.0
    init_min: float = 0.0
    init_max: float = 0.0
    clip_min: float = -jnp.finfo(jnp.float32).max
    clip_max: float = jnp.finfo(jnp.float32).max


def get_elite_weights(elite_popsize: int) -> Tuple[chex.Array, chex.Array]:
    """Utility helper to create truncated elite weights for mean update."""
    weights = jnp.array(
        [
            (
                (jnp.log(elite_popsize + 1) - jnp.log(i + 1))
                / (
                    elite_popsize * jnp.log(elite_popsize + 1)
                    - jnp.sum(jnp.log(jnp.arange(1, elite_popsize + 1)))
                )
            )
            for i in range(elite_popsize)
        ]
    )
    return weights


class RmES(Strategy):
    def __init__(
        self,
        popsize: int,
        num_dims: Optional[int] = None,
        pholder_params: Optional[Union[chex.ArrayTree, chex.Array]] = None,
        elite_ratio: float = 0.5,
        memory_size: int = 10,
        sigma_init: float = 1.0,
        mean_decay: float = 0.0,
        n_devices: Optional[int] = None,
        **fitness_kwargs: Union[bool, int, float]
    ):
        """Rank-m ES (Li & Zhang, 2017)
        Reference: https://ieeexplore.ieee.org/document/8080257
        """
        super().__init__(
            popsize,
            num_dims,
            pholder_params,
            mean_decay,
            n_devices,
            **fitness_kwargs
        )
        assert 0 <= elite_ratio <= 1
        self.elite_ratio = elite_ratio
        self.elite_popsize = max(1, int(self.popsize * self.elite_ratio))
        self.memory_size = memory_size  # number of ranks
        self.strategy_name = "RmES"

        # Set core kwargs es_params
        self.sigma_init = sigma_init

    @property
    def params_strategy(self) -> EvoParams:
        """Return default parameters of evolution strategy."""
        weights = get_elite_weights(self.elite_popsize)
        mu_eff = 1 / jnp.sum(weights ** 2)
        c_cov = 1 / (3 * jnp.sqrt(self.num_dims) + 5)
        c_c = 2 / (self.num_dims + 7)
        params = EvoParams(
            c_cov=c_cov,
            c_c=c_c,
            c_sigma=jnp.minimum(2 / (self.num_dims + 7), 0.05),
            mu_eff=mu_eff,
            sigma_init=self.sigma_init,
        )
        return params

    def initialize_strategy(
        self, rng: chex.PRNGKey, params: EvoParams
    ) -> EvoState:
        """`initialize` the evolution strategy."""
        weights = get_elite_weights(self.elite_popsize)
        # Initialize evolution paths & covariance matrix
        initialization = jax.random.uniform(
            rng,
            (self.num_dims,),
            minval=params.init_min,
            maxval=params.init_max,
        )
        state = EvoState(
            p_sigma=jnp.zeros(self.num_dims),
            sigma=params.sigma_init,
            mean=initialization,
            P=jnp.zeros((self.num_dims, self.memory_size)),
            t_gap=jnp.zeros(self.memory_size),
            s_rank_rate=0.0,
            weights=weights,
            # Store previous generations fitness for rank-based success rule
            fitness_archive=jnp.zeros(self.popsize) + 1e20,
            best_member=initialization,
        )
        return state

    def ask_strategy(
        self, rng: chex.PRNGKey, state: EvoState, params: EvoParams
    ) -> Tuple[chex.Array, EvoState]:
        """`ask` for new parameter candidates to evaluate next."""
        x = sample(
            rng,
            state.mean,
            state.sigma,
            state.P,
            self.num_dims,
            self.popsize,
            params.c_cov,
            state.gen_counter,
        )
        return x, state

    def tell_strategy(
        self,
        x: chex.Array,
        fitness: chex.Array,
        state: EvoState,
        params: EvoParams,
    ) -> EvoState:
        """`tell` performance data for strategy state update."""
        # Sort new results, extract elite, store best performer
        concat_p_f = jnp.hstack([jnp.expand_dims(fitness, 1), x])
        sorted_solutions = concat_p_f[concat_p_f[:, 0].argsort()][
            : self.elite_popsize
        ]
        # Update mean, isotropic/anisotropic paths, covariance, stepsize
        mean = update_mean(
            state.mean, sorted_solutions, params.c_m, state.weights
        )
        p_sigma = update_p_sigma(
            state.mean,
            mean,
            state.sigma,
            state.p_sigma,
            params.c_sigma,
            params.mu_eff,
        )
        P, t_gap = update_P_matrix(
            state.P,
            state.p_sigma,
            state.t_gap,
            params.t_uncorr,
            state.gen_counter,
        )

        s_rank_rate = rank_success_rule(
            fitness,
            state.fitness_archive,
            state.s_rank_rate,
            params.q_star,
            state.weights,
            params.c_s,
        )
        sigma = update_sigma(state.sigma, s_rank_rate, params.d_sigma)
        sigma = jnp.maximum(sigma, params.sigma_limit)
        return state.replace(
            mean=mean,
            p_sigma=p_sigma,
            P=P,
            t_gap=t_gap,
            sigma=sigma,
            s_rank_rate=s_rank_rate,
            fitness_archive=fitness,
        )


def update_mean(
    mean: chex.Array,
    sorted_solutions: chex.Array,
    c_m: float,
    weights: chex.Array,
) -> chex.Array:
    """Update mean of strategy."""
    mean = (1 - c_m) * mean + c_m * jnp.sum(
        sorted_solutions[:, 1:].T * weights, axis=1
    )
    return mean


def update_p_sigma(
    mean_old: chex.Array,
    mean: chex.Array,
    sigma: float,
    p_sigma: chex.Array,
    c_sigma: float,
    mu_eff: float,
) -> chex.Array:
    """Update evolution path for covariance matrix."""
    p_sigma_new = (1 - c_sigma) * p_sigma + jnp.sqrt(
        c_sigma * (2 - c_sigma) * mu_eff
    ) * (mean - mean_old) / sigma
    return p_sigma_new


def update_P_matrix(
    P: chex.Array,
    p_sigma: chex.Array,
    t_gap: chex.Array,
    t_uncorr: int,
    gen_counter: int,
) -> Tuple[chex.Array, chex.Array]:
    """Update the P matrix storing m evolution paths."""
    memory_size = P.shape[1]
    # Use evo paths in separated generations - keep them uncorrelated!
    T_min = jnp.min(t_gap[1:] - t_gap[:-1])
    replace_crit = T_min > t_uncorr
    fill_up_crit = gen_counter < memory_size
    push_replace = jnp.logical_or(replace_crit, fill_up_crit)

    # Case 1: Initially Record all evolution paths - make space for new one
    P_c1 = P.at[:, :-1].set(P[:, 1:])
    t_gap_c1 = t_gap.at[:-1].set(t_gap[1:])

    # Case 2: Remove/overwrite the oldesy recorded evolution path
    # Problem: i_min is a dynamic index - write as sum of two matrices?!
    P_c2 = P[:]
    t_gap_c2 = t_gap[:]
    i_min = jnp.argmin(t_gap[:-1] - t_gap[1:])
    for i in range(memory_size - 1):
        replace_bool = i >= i_min
        P_c2 = jax.lax.select(
            replace_bool, P_c2.at[:, i].set(P_c2[:, i + 1]), P_c2
        )
        t_gap_c2 = jax.lax.select(
            replace_bool, t_gap_c2.at[i].set(t_gap_c2[i + 1]), t_gap_c2
        )

    P = jax.lax.select(push_replace, P_c1, P_c2)
    t_gap = jax.lax.select(push_replace, t_gap_c1, t_gap_c2)

    # Finally update with the most recent evolution path
    P = P.at[:, memory_size - 1].set(p_sigma)
    t_gap = t_gap.at[memory_size - 1].set(gen_counter)
    return P, t_gap


def update_sigma(sigma: float, s_rank_rate: float, d_sigma: float) -> float:
    """Update stepsize sigma."""
    sigma_new = sigma * jnp.exp(s_rank_rate / d_sigma)
    return sigma_new


def sample(
    rng: chex.PRNGKey,
    mean: chex.Array,
    sigma: float,
    P: chex.Array,
    n_dim: int,
    pop_size: int,
    c_cov: chex.Array,
    gen_counter: int,
) -> chex.Array:
    """Jittable Gaussian Sample Helper."""
    z = jax.random.normal(rng, (n_dim, pop_size))  # ~ N(0, I)
    r = jax.random.normal(rng, (n_dim, P.shape[1]))

    for j in range(P.shape[1]):
        update_bool = gen_counter > j
        new_z = (
            jnp.sqrt(1 - c_cov) * z
            + (jnp.sqrt(c_cov) * P[:, j])[:, jnp.newaxis]
            * r[:, j][:, jnp.newaxis]
        )
        z = jax.lax.select(update_bool, new_z, z)
    z = jnp.swapaxes(z, 1, 0)
    x = mean + sigma * z  # ~ N(m, σ^2 C)
    return x


def rank_success_rule(
    fitness: chex.Array,
    fitness_archive: chex.Array,
    s_rank_rate: float,
    q_star: float,
    weights: chex.Array,
    c_s: float,
) -> float:
    """Compute rank-based success rule (cumulative rank rate)."""
    elite_popsize = weights.shape[0]
    popsize = fitness.shape[0]

    # Step 1: Sort all fitnesses in ascending order and get ranks
    # Rank parents + kids jointly - subdivide afterwards & take elite from both
    concat_all = jnp.vstack(
        [jnp.expand_dims(fitness, 1), jnp.expand_dims(fitness_archive, 1)]
    )
    ranks = jnp.zeros(concat_all.shape[0])
    ranks = ranks.at[concat_all[:, 0].argsort()].set(jnp.arange(2 * popsize))

    ranks_current = ranks[:popsize]
    ranks_current = ranks_current[ranks_current.argsort()][:elite_popsize]
    ranks_last = ranks[popsize:]
    ranks_last = ranks_last[ranks_last.argsort()][:elite_popsize]

    # Step 2: Compute rank difference (Parents vs. kids) - paper assumes min!
    q = 1 / elite_popsize * jnp.sum(weights * (ranks_last - ranks_current))

    # Step 3: Compute comulative rank rate using decaying memory
    new_s_rank_rate = (1 - c_s) * s_rank_rate + c_s * (q - q_star)
    return new_s_rank_rate



from typing import Tuple, Optional, Union
import jax
import jax.numpy as jnp
import chex
from flax import struct
from ..strategy import Strategy


@struct.dataclass
class EvoState:
    mean: chex.Array
    best_member: chex.Array
    best_fitness: float = jnp.finfo(jnp.float32).max
    gen_counter: int = 0


@struct.dataclass
class EvoParams:
    init_min: float = 0.0
    init_max: float = 0.0
    range_min: float = 0.0
    range_max: float = 1.0
    clip_min: float = -jnp.finfo(jnp.float32).max
    clip_max: float = jnp.finfo(jnp.float32).max


class RandomSearch(Strategy):
    def __init__(
        self,
        popsize: int,
        num_dims: Optional[int] = None,
        pholder_params: Optional[Union[chex.ArrayTree, chex.Array]] = None,
        n_devices: Optional[int] = None,
        **fitness_kwargs: Union[bool, int, float]
    ):
        """Simple Random Search Baseline"""

        super().__init__(
            popsize,
            num_dims,
            pholder_params,
            n_devices=n_devices,
            **fitness_kwargs
        )
        self.strategy_name = "RandomSearch"

    @property
    def params_strategy(self) -> EvoParams:
        """Return default parameters of evolution strategy."""
        return EvoParams()

    def initialize_strategy(
        self, rng: chex.PRNGKey, params: EvoParams
    ) -> EvoState:
        """`initialize` the differential evolution strategy."""
        initialization = jax.random.uniform(
            rng,
            (self.num_dims,),
            minval=params.init_min,
            maxval=params.init_max,
        )
        state = EvoState(
            mean=initialization,
            best_member=initialization,
        )
        return state

    def ask_strategy(
        self, rng: chex.PRNGKey, state: EvoState, params: EvoParams
    ) -> Tuple[chex.Array, EvoState]:
        """`ask` for new proposed candidates to evaluate next."""
        x = jax.random.uniform(
            rng,
            (self.popsize, self.num_dims),
            minval=params.range_min,
            maxval=params.range_max,
        )
        return x, state

    def tell_strategy(
        self,
        x: chex.Array,
        fitness: chex.Array,
        state: EvoState,
        params: EvoParams,
    ) -> EvoState:
        """`tell` update to ES state."""
        idx = jnp.argsort(fitness)
        fitness = fitness[idx]
        x = x[idx]
        # Set mean to best member seen so far
        improved = fitness[0] < state.best_fitness
        best_mean = jax.lax.select(improved, x[0], state.best_member)
        return state.replace(mean=best_mean)



from flax import linen as nn
import chex
from typing import Optional
from .shared import (
    identity_out,
    tanh_out,
    categorical_out,
    gaussian_out,
    default_bias_init,
    kernel_init_fn,
)


class MLP(nn.Module):
    """Simple MLP Wrapper with flexible output head."""

    num_hidden_units: int = 64
    num_hidden_layers: int = 2
    num_output_units: int = 1
    hidden_activation: str = "relu"
    output_activation: str = "identity"
    kernel_init_type: str = "lecun_normal"
    model_name: str = "MLP"

    @nn.compact
    def __call__(
        self, x: chex.Array, rng: Optional[chex.PRNGKey] = None
    ) -> chex.Array:
        # Flatten a single 3d image into a plain flat vector
        if len(x.shape) <= 3:
            x = x.reshape(-1)
        # Flatten a batch of 3d images into a batch of flat vectors
        else:
            x = x.reshape(x.shape[0], -1)

        # Loop over dense layers in forward pass
        for _ in range(self.num_hidden_layers):
            x = nn.Dense(
                features=self.num_hidden_units,
                kernel_init=kernel_init_fn[self.kernel_init_type](),
                bias_init=default_bias_init(),
            )(x)
            if self.hidden_activation == "relu":
                x = nn.relu(x)
            elif self.hidden_activation == "tanh":
                x = nn.tanh(x)
            elif self.hidden_activation == "gelu":
                x = nn.gelu(x)
            elif self.hidden_activation == "softplus":
                x = nn.softplus(x)

        if self.output_activation == "identity":
            return identity_out(x, self.num_output_units, self.kernel_init_type)
        elif self.output_activation == "tanh":
            return tanh_out(x, self.num_output_units, self.kernel_init_type)
        # Categorical and gaussian output heads require rng for sampling
        elif self.output_activation == "categorical":
            return categorical_out(
                rng, x, self.num_output_units, self.kernel_init_type
            )
        elif self.output_activation == "gaussian":
            return gaussian_out(
                rng, x, self.num_output_units, self.kernel_init_type
            )



from typing import Union, Optional
import jax
import jax.numpy as jnp
import chex
from functools import partial


class FitnessShaper(object):
    def __init__(
        self,
        centered_rank: Union[bool, int] = False,
        z_score: Union[bool, int] = False,
        norm_range: Union[bool, int] = False,
        w_decay: float = 0.0,
        maximize: Union[bool, int] = False,
        fitness_trafo: Optional[str] = None,
    ):
        """JAX-compatible fitness shaping tool."""
        self.w_decay = w_decay
        self.maximize = bool(maximize)

        if fitness_trafo in ["centered_rank", "z_score", "norm_range", "raw"]:
            self.centered_rank = fitness_trafo == "centered_rank"
            self.z_score = fitness_trafo == "z_score"
            self.norm_range = fitness_trafo == "norm_range"
        else:
            self.centered_rank = bool(centered_rank)
            self.z_score = bool(z_score)
            self.norm_range = bool(norm_range)
        # Check that only single fitness shaping transformation is used
        num_options_on = self.centered_rank + self.z_score + self.norm_range
        assert (
            num_options_on < 2
        ), "Only use one fitness shaping transformation."

    @partial(jax.jit, static_argnums=(0,))
    def apply(self, x: chex.Array, fitness: chex.Array) -> chex.Array:
        """Max objective trafo, rank shaping, z scoring & add weight decay."""
        if self.maximize:
            fitness = -1 * fitness

        # Apply wdecay before normalization - makes easier to tune
        # "Reduce" fitness based on L2 norm of parameters
        if self.w_decay > 0.0:
            l2_fit_red = self.w_decay * compute_l2_norm(x)
            fitness += l2_fit_red

        if self.centered_rank:
            fitness = centered_rank_trafo(fitness)

        if self.z_score:
            fitness = z_score_trafo(fitness)

        if self.norm_range:
            fitness = range_norm_trafo(fitness, -1.0, 1.0)

        return fitness


def z_score_trafo(arr: chex.Array) -> chex.Array:
    """Make fitness 'Gaussian' by substracting mean and dividing by std."""
    return (arr - jnp.nanmean(arr)) / (jnp.nanstd(arr) + 1e-10)


def compute_ranks(fitness: chex.Array) -> chex.Array:
    """Return fitness ranks in [0, len(fitness))."""
    ranks = jnp.zeros(len(fitness))
    ranks = ranks.at[fitness.argsort()].set(jnp.arange(len(fitness)))
    return ranks


def centered_rank_trafo(fitness: chex.Array) -> chex.Array:
    """Return ~ -0.5 to 0.5 centered ranks (best to worst - min!)."""
    y = compute_ranks(fitness)
    y /= fitness.size - 1
    return y - 0.5


def compute_l2_norm(x: chex.Array) -> chex.Array:
    """Compute L2-norm of x_i. Assumes x to have shape (popsize, num_dims)."""
    return jnp.nanmean(x * x, axis=1)


def range_norm_trafo(
    arr: chex.Array, min_val: float = -1.0, max_val: float = 1.0
) -> chex.Array:
    """Map scores into a min/max range."""
    arr = jnp.clip(arr, -1e10, 1e10)
    normalized_arr = (max_val - min_val) * (arr - jnp.nanmin(arr)) / (
        jnp.nanmax(arr) - jnp.nanmin(arr) + 1e-10
    ) + min_val
    return normalized_arr



import chex
import jax.numpy as jnp
from typing import Tuple


def full_eigen_decomp(
    C: chex.Array, B: chex.Array, D: chex.Array, gen_counter: int
) -> Tuple[chex.Array, chex.Array, chex.Array]:
    """Perform eigendecomposition of covariance matrix."""
    if B is not None and D is not None:
        return C, B, D
    C = C + 1e-10 * (gen_counter == 0)
    C = (C + C.T) / 2  # Make sure matrix is symmetric
    D2, B = jnp.linalg.eigh(C)
    D = jnp.sqrt(jnp.where(D2 < 0, 1e-20, D2))
    C = jnp.dot(jnp.dot(B, jnp.diag(D ** 2)), B.T)
    return C, B, D


def diag_eigen_decomp(C: chex.Array, D: chex.Array) -> chex.Array:
    """Perform simplified decomposition of diagonal covariance matrix."""
    if D is not None:
        return D
    D = jnp.sqrt(jnp.where(C < 0, 1e-20, C))
    return D



from typing import Tuple, Optional, Union
import jax
import chex
from functools import partial
from flax import struct
from .cma_es import CMA_ES
from ..restarts.restarter import WrapperState, WrapperParams


@struct.dataclass
class RestartParams:
    min_num_gens: int = 50
    min_fitness_spread: float = 1e-12
    popsize_multiplier: int = 2
    tol_x: float = 1e-12
    tol_x_up: float = 1e4
    tol_condition_C: float = 1e14
    copy_mean: bool = True


class BIPOP_CMA_ES(object):
    def __init__(
        self,
        popsize: int,
        num_dims: Optional[int] = None,
        pholder_params: Optional[Union[chex.ArrayTree, chex.Array]] = None,
        elite_ratio: float = 0.5,
        sigma_init: float = 1.0,
        mean_decay: float = 0.0,
        n_devices: Optional[int] = None,
        **fitness_kwargs: Union[bool, int, float]
    ):
        """BIPOP-CMA-ES (Hansen, 2009).
        Reference: https://hal.inria.fr/inria-00382093/document
        Inspired by: https://tinyurl.com/44y3ryhf"""
        self.strategy_name = "BIPOP_CMA_ES"
        # Instantiate base strategy & wrap it with restart wrapper
        self.strategy = CMA_ES(
            num_dims=num_dims,
            popsize=popsize,
            pholder_params=pholder_params,
            elite_ratio=elite_ratio,
            sigma_init=sigma_init,
            mean_decay=mean_decay,
            n_devices=n_devices,
            **fitness_kwargs,
        )
        from ..restarts import BIPOP_Restarter
        from ..restarts.termination import spread_criterion, cma_criterion

        self.wrapped_strategy = BIPOP_Restarter(
            self.strategy,
            stop_criteria=[spread_criterion, cma_criterion],
            strategy_kwargs={
                "elite_ratio": elite_ratio,
                "mean_decay": mean_decay,
            },
        )

    @property
    def default_params(self) -> WrapperParams:
        """Return default parameters of evolution strategy."""
        re_params = self.wrapped_strategy.default_params
        return re_params.replace(restart_params=RestartParams())

    @partial(jax.jit, static_argnums=(0,))
    def initialize(
        self, rng: chex.PRNGKey, params: Optional[WrapperParams] = None
    ) -> WrapperState:
        """`initialize` the evolution strategy."""
        # Use default hyperparameters if no other settings provided
        if params is None:
            params = self.default_params
        return self.wrapped_strategy.initialize(rng, params)

    def ask(
        self,
        rng: chex.PRNGKey,
        state: WrapperState,
        params: Optional[WrapperParams] = None,
    ) -> Tuple[chex.Array, WrapperState]:
        """`ask` for new parameter candidates to evaluate next."""
        # Use default hyperparameters if no other settings provided
        if params is None:
            params = self.default_params
        x, state = self.wrapped_strategy.ask(rng, state, params)
        return x, state

    @partial(jax.jit, static_argnums=(0,))
    def tell(
        self,
        x: chex.Array,
        fitness: chex.Array,
        state: WrapperState,
        params: Optional[WrapperParams] = None,
    ) -> WrapperState:
        """`tell` performance data for strategy state update."""
        # Use default hyperparameters if no other settings provided
        if params is None:
            params = self.default_params
        return self.wrapped_strategy.tell(x, fitness, state, params)



from typing import Tuple, Optional, Union
import jax
import jax.numpy as jnp
import chex
from flax import struct
from ..strategy import Strategy
from ..core import exp_decay
from .full_iamalgam import (
    anticipated_mean_shift,
    adaptive_variance_scaling,
    update_mean_amalgam,
)


@struct.dataclass
class EvoState:
    mean: chex.Array
    mean_shift: chex.Array
    C: chex.Array
    nis_counter: int
    c_mult: float
    sigma: float
    best_member: chex.Array
    best_fitness: float = jnp.finfo(jnp.float32).max
    gen_counter: int = 0


@struct.dataclass
class EvoParams:
    eta_sigma: float
    eta_shift: float
    eta_avs_inc: float = 1.0 / 0.9
    eta_avs_dec: float = 0.9
    nis_max_gens: int = 50
    delta_ams: float = 2.0
    theta_sdr: float = 1.0
    c_mult_init: float = 1.0
    sigma_init: float = 0.0
    sigma_decay: float = 0.999
    sigma_limit: float = 0.0
    init_min: float = 0.0
    init_max: float = 0.0
    clip_min: float = -jnp.finfo(jnp.float32).max
    clip_max: float = jnp.finfo(jnp.float32).max


class Indep_iAMaLGaM(Strategy):
    def __init__(
        self,
        popsize: int,
        num_dims: Optional[int] = None,
        pholder_params: Optional[Union[chex.ArrayTree, chex.Array]] = None,
        elite_ratio: float = 0.35,
        sigma_init: float = 0.0,
        sigma_decay: float = 0.99,
        sigma_limit: float = 0.0,
        mean_decay: float = 0.0,
        n_devices: Optional[int] = None,
        **fitness_kwargs: Union[bool, int, float]
    ):
        """(Iterative) AMaLGaM (Bosman et al., 2013) - Diagonal Covariance
        Reference: https://tinyurl.com/y9fcccx2
        """
        super().__init__(
            popsize,
            num_dims,
            pholder_params,
            mean_decay,
            n_devices,
            **fitness_kwargs
        )
        assert 0 <= elite_ratio <= 1
        self.elite_ratio = elite_ratio
        self.elite_popsize = max(1, int(self.popsize * self.elite_ratio))
        alpha_ams = (
            0.5
            * self.elite_ratio
            * self.popsize
            / (self.popsize - self.elite_popsize)
        )
        self.ams_popsize = int(alpha_ams * (self.popsize - 1))
        self.strategy_name = "Indep_iAMaLGaM"

        # Set core kwargs es_params
        self.sigma_init = sigma_init
        self.sigma_decay = sigma_decay
        self.sigma_limit = sigma_limit

    @property
    def params_strategy(self) -> EvoParams:
        """Return default parameters of evolution strategy."""
        a_0_sigma, a_1_sigma, a_2_sigma = -1.1, 1.2, 1.6
        a_0_shift, a_1_shift, a_2_shift = -1.2, 0.31, 0.5
        eta_sigma = 1 - jnp.exp(
            a_0_sigma
            * self.elite_popsize ** a_1_sigma
            / (self.num_dims ** a_2_sigma)
        )
        eta_shift = 1 - jnp.exp(
            a_0_shift
            * self.elite_popsize ** a_1_shift
            / (self.num_dims ** a_2_shift)
        )

        return EvoParams(
            eta_sigma=eta_sigma,
            eta_shift=eta_shift,
            sigma_init=self.sigma_init,
            sigma_decay=self.sigma_decay,
            sigma_limit=self.sigma_limit,
        )

    def initialize_strategy(
        self, rng: chex.PRNGKey, params: EvoParams
    ) -> EvoState:
        """`initialize` the evolution strategy."""
        # Initialize evolution paths & covariance matrix
        initialization = jax.random.uniform(
            rng,
            (self.num_dims,),
            minval=params.init_min,
            maxval=params.init_max,
        )
        state = EvoState(
            mean=initialization,
            mean_shift=jnp.zeros(self.num_dims),
            C=jnp.ones(self.num_dims),
            sigma=params.sigma_init,
            nis_counter=0,
            c_mult=params.c_mult_init,
            best_member=initialization,
        )
        return state

    def ask_strategy(
        self, rng: chex.PRNGKey, state: EvoState, params: EvoParams
    ) -> Tuple[chex.Array, EvoState]:
        """`ask` for new parameter candidates to evaluate next."""
        rng_sample, rng_ams = jax.random.split(rng)
        x = sample(rng_sample, state.mean, state.C, state.sigma, self.popsize)
        x_ams = anticipated_mean_shift(
            rng_ams,
            x,
            self.ams_popsize,
            params.delta_ams,
            state.c_mult,
            state.mean_shift,
        )
        return x_ams, state

    def tell_strategy(
        self,
        x: chex.Array,
        fitness: chex.Array,
        state: EvoState,
        params: EvoParams,
    ) -> EvoState:
        """`tell` performance data for strategy state update."""
        # Sort new results, extract elite
        idx = jnp.argsort(fitness)[0 : self.elite_popsize]
        fitness_elite = fitness[idx]
        members_elite = x[idx]

        # If there has been a fitness improvement -> Run AVS based on SDR
        improvements = fitness_elite < state.best_fitness
        any_improvement = jnp.sum(improvements) > 0
        sdr = standard_deviation_ratio(
            improvements, members_elite, state.mean, state.C
        )
        c_mult, nis_counter = adaptive_variance_scaling(
            any_improvement,
            sdr,
            state.c_mult,
            state.nis_counter,
            params.theta_sdr,
            params.eta_avs_inc,
            params.eta_avs_dec,
            params.nis_max_gens,
        )

        # Update mean and covariance estimates - difference full vs. indep
        mean, mean_shift = update_mean_amalgam(
            members_elite,
            state.mean,
            state.mean_shift,
            params.eta_shift,
        )
        C = update_cov_amalgam(members_elite, state.C, mean, params.eta_sigma)

        # Decay isotropic part of Gaussian search distribution
        sigma = exp_decay(state.sigma, params.sigma_decay, params.sigma_limit)
        return state.replace(
            c_mult=c_mult,
            nis_counter=nis_counter,
            mean=mean,
            mean_shift=mean_shift,
            C=C,
            sigma=sigma,
        )


def sample(
    rng: chex.PRNGKey,
    mean: chex.Array,
    C: chex.Array,
    sigma: float,
    popsize: int,
) -> chex.Array:
    """Jittable Gaussian Sample Helper."""
    sigmas = jnp.sqrt(C) + sigma
    z = jax.random.normal(rng, (mean.shape[0], popsize))  # ~ N(0, I)
    y = jnp.diag(sigmas).dot(z)  # ~ N(0, C)
    y = jnp.swapaxes(y, 1, 0)
    x = mean + y  # ~ N(m, σ^2 C)
    return x


def standard_deviation_ratio(
    improvements: chex.Array,
    members_elite: chex.Array,
    mean: chex.Array,
    C: chex.Array,
) -> float:
    """SDR - relate dist. of improvements to mean in param space."""
    # Compute avg. member for candidates that improve fitness -> SDR
    x_avg_imp = jnp.sum(
        improvements[:, jnp.newaxis] * members_elite, axis=0
    ) / jnp.sum(improvements)
    conditioned_diff = (x_avg_imp - mean) / C
    sdr = jnp.max(jnp.abs(conditioned_diff))
    return sdr


def update_cov_amalgam(
    members_elite: chex.Array,
    C: chex.Array,
    mean: chex.Array,
    eta_sigma: float,
) -> chex.Array:
    """Iterative update of mean and mean shift based on elite and history."""
    S_bar = members_elite - mean
    # Univariate update to standard deviations
    new_C = (1 - eta_sigma) * C + eta_sigma * jnp.sum(
        S_bar ** 2, axis=0
    ) / members_elite.shape[0]
    return new_C



import jax
import jax.numpy as jnp
import numpy as np
from evosax.experimental.subpops import BatchStrategy, MetaStrategy


def test_batch_strategy():
    rng = jax.random.PRNGKey(0)

    num_dims = 2
    popsize = 100
    num_subpops = 5

    batch_strategy = BatchStrategy(
        strategy_name="CMA_ES",
        num_dims=num_dims,
        popsize=popsize,
        num_subpops=num_subpops,
        strategy_kwargs={"elite_ratio": 0.5},
    )
    es_params = batch_strategy.default_params
    state = batch_strategy.initialize(rng, es_params)
    assert state.mean.shape == (num_subpops, num_dims)

    x, state = batch_strategy.ask(rng, state, es_params)
    assert x.shape == (popsize, num_dims)

    fitness = jnp.zeros(popsize)
    state = batch_strategy.tell(x, fitness, state, es_params)
    assert state.mean.shape == (num_subpops, num_dims)


def test_meta_strategy():
    rng = jax.random.PRNGKey(0)

    num_dims = 2
    popsize = 100
    num_subpops = 10

    meta_strategy = MetaStrategy(
        meta_strategy_name="CMA_ES",
        inner_strategy_name="DE",
        meta_params=["diff_w", "cross_over_rate"],
        num_dims=num_dims,
        popsize=popsize,
        num_subpops=num_subpops,
        meta_strategy_kwargs={"elite_ratio": 0.5},
    )
    meta_es_params = meta_strategy.default_params_meta.replace(
        clip_min=jnp.array([0, 0]), clip_max=jnp.array([2, 1])
    )

    # META: Initialize the meta strategy state
    inner_es_params = meta_strategy.default_params
    meta_state = meta_strategy.initialize_meta(rng, meta_es_params)

    # META: Get altered inner es hyperparams
    inner_es_params, meta_state = meta_strategy.ask_meta(
        rng, meta_state, meta_es_params, inner_es_params
    )
    assert meta_state.mean.shape == (num_dims,)

    # INNER: Initialize the inner batch ES
    state = meta_strategy.initialize(rng, inner_es_params)
    assert state.mean.shape == (num_subpops, num_dims)

    # INNER: Ask for inner candidate params to evaluate on problem
    x, state = meta_strategy.ask(rng, state, inner_es_params)
    assert x.shape == (popsize, num_dims)

    # INNER: Update using pseudo fitness
    fitness = jax.random.normal(rng, (popsize,))
    state = meta_strategy.tell(x, fitness, state, inner_es_params)
    assert state.mean.shape == (num_subpops, num_dims)

    # META: Update the meta strategy
    meta_state = meta_strategy.tell_meta(
        inner_es_params, fitness, meta_state, meta_es_params
    )
    assert meta_state.mean.shape == (num_dims,)


def test_protocol_best_subpop_strategy():
    rng = jax.random.PRNGKey(0)
    batch_strategy = BatchStrategy(
        strategy_name="CMA_ES",
        num_dims=2,
        popsize=100,
        num_subpops=5,
        strategy_kwargs={"elite_ratio": 0.5},
        communication="best_subpop",
    )
    es_params = batch_strategy.default_params
    state = batch_strategy.initialize(rng, es_params)
    assert state.mean.shape == (5, 2)

    x, state = batch_strategy.ask(rng, state, es_params)
    assert x.shape == (100, 2)

    fitness = jnp.ones(100)

    best_ind = np.random.randint(100)
    best_fitness = 0
    fitness = fitness.at[best_ind].set(best_fitness)

    state = batch_strategy.tell(x, fitness, state, es_params)

    assert state.mean.shape == (5, 2)



import jax.numpy as jnp
from evosax.utils import ESLog


def test_es_log():
    es_logging = ESLog(num_dims=2, num_generations=10, top_k=3, maximize=True)
    log = es_logging.initialize()
    x = jnp.array([[1, 2], [2, 4], [4, 6], [6, 7]])
    fitness = jnp.array([1, 2, 3, 4])
    log = es_logging.update(log, x, fitness)
    assert log["log_top_1"][0] == 4
    assert log["top_params"][0][0] == 6
    assert log["top_params"][0][1] == 7

def test_top_k():
    es_logging = ESLog(num_dims=2, num_generations=10, top_k=3, maximize=True)
    log = es_logging.initialize()
    x = jnp.array([[1, 2], [2, 4], [4, 6], [6, 7]])
    fitness = jnp.array([1, 2, 3, 4])
    log = es_logging.update(log, x, fitness)
    assert jnp.array_equal(log["top_fitness"], jnp.array([4, 3, 2]))
    assert jnp.array_equal(log["top_params"][0], jnp.array([6, 7]))
    assert jnp.array_equal(log["top_params"][1], jnp.array([4, 6]))
    assert jnp.array_equal(log["top_params"][2], jnp.array([2, 4]))



"""Fitness landscape visualizer and evaluation animator."""
import chex
import jax
import jax.numpy as jnp
import numpy as np
import matplotlib.cm as cm
import matplotlib.pyplot as plt
import matplotlib.animation as animation

# cmap = cm.colors.LinearSegmentedColormap.from_list(
#     "Custom", [(0, "#2f9599"), (0.45, "#eee"), (1, "#8800ff")], N=256
# )

cmap = cm.colors.LinearSegmentedColormap.from_list(
    "Custom", [(0, "#992f2f"), (0.45, "#eee"), (1, "#3a2f99")], N=256
)


class BBOBVisualizer(object):
    """Fitness landscape visualizer and evaluation animator."""

    def __init__(
        self,
        X: chex.Array,
        fitness: chex.Array,
        fn_name: str = "Rastrigin",
        title: str = "",
        use_3d: bool = False,
        plot_log_fn: bool = False,
        seed_id: int = 1,
        interval: int = 50,
        plot_title: bool = True,
        plot_labels: bool = True,
        plot_colorbar: bool = True,
    ):
        from evosax.problems.bbob import BBOB_fns, get_rotation

        self.X = X
        self.fitness = fitness
        self.title = title
        self.fn_name = fn_name
        self.use_3d = use_3d
        if not self.use_3d:
            self.fig, self.ax = plt.subplots(figsize=(6, 5))
        else:
            self.fig = plt.figure(figsize=(6, 5))
            self.ax = self.fig.add_subplot(1, 1, 1, projection="3d")
        self.fn_name = fn_name
        self.fn = BBOB_fns[self.fn_name]

        rng = jax.random.PRNGKey(seed_id)
        rng_q, rng_r = jax.random.split(rng)
        self.R = get_rotation(rng_r, 2)
        self.Q = get_rotation(rng_q, 2)
        self.global_minima = []

        # Set plot configuration
        self.plot_log_fn = plot_log_fn
        self.plot_title = plot_title
        self.plot_labels = plot_labels
        self.plot_colorbar = plot_colorbar

        # Set boundaries for evaluation range of black-box functions
        self.x1_lower_bound, self.x1_upper_bound = -5, 5
        self.x2_lower_bound, self.x2_upper_bound = -5, 5

        # Set meta-data for rotation/azimuth
        self.interval = interval  # Delay between frames in milliseconds.
        try:
            self.num_frames = X.shape[0]
            self.static_frames = int(0.2 * self.num_frames)
            self.azimuths = jnp.linspace(
                0, 89, self.num_frames - self.static_frames
            )
            self.angles = jnp.linspace(
                0, 89, self.num_frames - self.static_frames
            )
        except Exception:
            pass

    def animate(self, save_fname: str):
        """Run animation for provided data."""
        ani = animation.FuncAnimation(
            self.fig,
            self.update,
            frames=self.num_frames,
            init_func=self.init,
            blit=False,
            interval=self.interval,
        )
        ani.save(save_fname)

    def init(self):
        """Initialize the first frame for the animation."""
        if self.use_3d:
            self.plot_contour_3d()
            (self.scat,) = self.ax.plot(
                self.X[0, :, 0],
                self.X[0, :, 1],
                self.fitness[0, :],
                marker="o",
                c="y",
                linestyle="",
                markersize=10,
                alpha=0.75,
            )

        else:
            self.plot_contour_2d()
            (self.scat,) = self.ax.plot(
                self.X[0, :, 0],
                self.X[0, :, 1],
                marker="o",
                c="y",
                linestyle="",
                markersize=10,
                alpha=0.75,
            )

        return (self.scat,)

    def update(self, frame):
        """Update the frame with the solutions evaluated in generation."""
        # Plot sample points
        self.scat.set_data(self.X[frame, :, 0], self.X[frame, :, 1])
        if self.use_3d:
            if self.plot_log_fn:
                fit = jnp.log(self.fitness[frame, :])
            else:
                fit = self.fitness[frame, :]
            self.scat.set_3d_properties(fit)
            if frame < self.num_frames - self.static_frames:
                self.ax.view_init(self.azimuths[frame], self.angles[frame])

        if self.plot_title:
            if self.plot_log_fn:
                self.ax.set_title(
                    f"Log {self.fn_name}: {self.title} - Generation"
                    f" {frame + 1}",
                    fontsize=15,
                )
            else:
                self.ax.set_title(
                    f"{self.fn_name}: {self.title} - Generation {frame + 1}",
                    fontsize=15,
                )
        self.fig.tight_layout()
        return (self.scat,)

    def contour_function(self, x1, x2):
        """Evaluate vmapped fitness landscape."""

        def fn_val(x1, x2):
            x = jnp.stack([x1, x2])
            return self.fn(x, self.R, self.Q)

        return jax.vmap(jax.vmap(fn_val, in_axes=(0, None)), in_axes=(None, 0))(
            x1, x2
        )

    def plot_contour_2d(self, save: bool = False):
        """Plot 2d landscape contour."""

        if save:
            self.fig, self.ax = plt.subplots(figsize=(6, 5))
        self.ax.set_xlim(self.x1_lower_bound, self.x1_upper_bound)
        self.ax.set_ylim(self.x2_lower_bound, self.x2_upper_bound)
        self.ax.set_xlim(self.x1_lower_bound, self.x1_upper_bound)
        self.ax.set_ylim(self.x2_lower_bound, self.x2_upper_bound)

        # Plot local minimum value
        for m in self.global_minima:
            self.ax.plot(m[0], m[1], "y*", ms=10)
            self.ax.plot(m[0], m[1], "y*", ms=10)

        x1 = jnp.arange(self.x1_lower_bound, self.x1_upper_bound, 0.01)
        x2 = jnp.arange(self.x2_lower_bound, self.x2_upper_bound, 0.01)
        X, Y = np.meshgrid(x1, x2)
        contour = self.contour_function(x1, x2)
        if self.plot_log_fn:
            contour = jnp.log(contour)
        self.ax.contour(X, Y, contour, levels=30, linewidths=0.5, colors="#999")
        im = self.ax.contourf(X, Y, contour, levels=30, cmap=cmap, alpha=0.7)
        if self.plot_title:
            if self.plot_log_fn:
                self.ax.set_title(f"Log {self.fn_name} Function", fontsize=15)
            else:
                self.ax.set_title(f"{self.fn_name} Function", fontsize=15)

        if self.plot_labels:
            self.ax.set_xlabel(r"$x_1$")
            self.ax.set_ylabel(r"$x_2$")
        else:
            self.ax.set_xticks([])
            self.ax.set_yticks([])

        if self.plot_colorbar:
            self.fig.colorbar(im, ax=self.ax)
        self.fig.tight_layout()

        if save:
            plt.savefig(f"{self.fn_name}_2d.png", dpi=300)

    def plot_contour_3d(self, save: bool = False):
        """Plot 3d landscape contour."""
        self.fig = plt.figure(figsize=(6, 5))
        self.ax = self.fig.add_subplot(1, 1, 1, projection="3d")
        x1 = jnp.arange(self.x1_lower_bound, self.x1_upper_bound, 0.01)
        x2 = jnp.arange(self.x2_lower_bound, self.x2_upper_bound, 0.01)
        contour = self.contour_function(x1, x2)
        if self.plot_log_fn:
            contour = jnp.log(contour)

        X, Y = np.meshgrid(x1, x2)
        self.ax.contour(
            X,
            Y,
            contour,
            zdir="z",
            offset=np.min(contour),
            levels=30,
            cmap=cmap,
            alpha=0.5,
        )
        self.ax.plot_surface(
            X,
            Y,
            contour,
            cmap=cmap,
            linewidth=0,
            antialiased=True,
            alpha=0.7,
        )

        # Rmove fills and set labels
        self.ax.xaxis.pane.fill = False
        self.ax.yaxis.pane.fill = False
        self.ax.zaxis.pane.fill = False

        self.ax.xaxis.set_tick_params(labelsize=8)
        self.ax.yaxis.set_tick_params(labelsize=8)
        self.ax.zaxis.set_tick_params(labelsize=8)

        if self.plot_labels:
            self.ax.set_xlabel(r"$x_1$")
            self.ax.set_ylabel(r"$x_2$")
        else:
            self.ax.set_xticks([])
            self.ax.set_yticks([])
            self.ax.set_zticks([])

        if self.plot_title:
            if self.plot_log_fn:
                self.ax.set_title(f"Log {self.fn_name} Function", fontsize=15)
                self.ax.set_zlabel(r"$\log f(x)$")
            else:
                self.ax.set_title(f"{self.fn_name} Function", fontsize=15)
                self.ax.set_zlabel(r"$f(x)$")
        self.fig.tight_layout()
        if save:
            plt.savefig(f"{self.fn_name}_3d.png", dpi=300)


if __name__ == "__main__":
    import jax
    from evosax import CMA_ES
    from evosax.problems import BBOBFitness

    # from jax.config import config
    # config.update("jax_enable_x64", True)

    # for fn_name in [
    #     "BuecheRastrigin",
    # ]:  # BBOB_fns.keys():
    #     print(f"Start 2d/3d - {fn_name}")
    #     visualizer = BBOBVisualizer(None, None, fn_name, "")
    #     visualizer.plot_contour_2d(save=True)
    #     visualizer.plot_contour_3d(save=True)

    # Test animations
    # All solutions from single run (10 gens, 16 pmembers, 2 dims)
    for fn_name in [
        "Sphere",
        "RosenbrockOriginal",
        # "RosenbrockRotated",
        "Discus",
        # "RastriginRotated",
        "Schwefel",
        # Large set of functions
        # "BuecheRastrigin",
        # "AttractiveSector",
        # "Weierstrass",
        "SchaffersF7",
        # "GriewankRosenbrock",
        # Part 1: Separable functions
        # "EllipsoidalOriginal",
        # "RastriginOriginal",
        # "LinearSlope",
        # Part 2: Functions with low or moderate conditions
        # "AttractiveSector",
        # "StepEllipsoidal",
        # Part 3: Functions with high conditioning and unimodal
        # "EllipsoidalRotated",
        "BentCigar",
        "SharpRidge",
        "DifferentPowers",
        # Part 4: Multi-modal functions with adequate global structure
        # "SchaffersF7IllConditioned",
        # Part 5: Multi-modal functions with weak global structure
        # "Lunacek",
        # "Gallagher101Me",
        # "Gallagher21Hi",
    ]:
        rng = jax.random.PRNGKey(1)
        strategy = CMA_ES(popsize=4, num_dims=2)
        es_params = strategy.default_params.replace(init_min=-2.5, init_max=2.5)
        es_state = strategy.initialize(rng, es_params)

        problem = BBOBFitness(fn_name, 2)

        X, fitness = [], []
        for g in range(50):
            rng, rng_ask, rng_eval = jax.random.split(rng, 3)
            x, es_state = strategy.ask(rng, es_state, es_params)
            fit = problem.rollout(rng_eval, x)
            es_state = strategy.tell(x, fit, es_state, es_params)
            X.append(x)
            fitness.append(fit)

        X = jnp.stack(X)
        fitness = jnp.stack(fitness)
        print(fn_name, fitness.shape, X.shape)
        visualizer = BBOBVisualizer(
            X,
            fitness,
            fn_name,
            "CMA-ES",
            use_3d=False,
            plot_log_fn=True,
            interval=100,
        )
        visualizer.animate(f"anims/{fn_name}_2d.gif")
        # visualizer = BBOBVisualizer(X, None, "Sphere", "Test Strategy", use_3d=False)
        # visualizer.animate("Sphere_2d.gif")



import jax
import chex
from functools import partial
from typing import Tuple, Optional
from .restarter import RestartWrapper, WrapperState, WrapperParams
from .termination import spread_criterion
from flax import struct


@struct.dataclass
class RestartState:
    restart_counter: int
    restart_next: bool
    active_popsize: int
    restart_large_counter: int
    large_eval_budget: int
    small_eval_budget: int
    small_pop_active: bool


@struct.dataclass
class RestartParams:
    min_num_gens: int = 50
    min_fitness_spread: float = 1e-12
    popsize_multiplier: int = 2
    copy_mean: bool = False


class BIPOP_Restarter(RestartWrapper):
    def __init__(
        self,
        base_strategy,
        stop_criteria=[spread_criterion],
        strategy_kwargs: dict = {},
    ):
        """Bi-Population Restarts (Hansen, 2009) - Interlaced population sizes.
        Reference: https://hal.inria.fr/inria-00382093/document
        Inspired by: https://tinyurl.com/44y3ryhf"""
        super().__init__(base_strategy, stop_criteria)
        self.default_popsize = self.base_strategy.popsize
        self.strategy_kwargs = strategy_kwargs

        from .. import Strategies

        global Strategies

    @property
    def restart_params(self) -> RestartParams:
        """Return default parameters for strategy restarting."""
        return RestartParams()

    @partial(jax.jit, static_argnums=(0,))
    def initialize(
        self, rng: chex.PRNGKey, params: Optional[WrapperParams] = None
    ) -> WrapperState:
        """`initialize` the evolution strategy."""
        # Use default hyperparameters if no other settings provided
        if params is None:
            params = self.default_params

        strategy_state = self.base_strategy.initialize(
            rng, params.strategy_params
        )
        restart_state = RestartState(
            restart_counter=0,
            restart_next=False,
            active_popsize=self.base_strategy.popsize,
            restart_large_counter=0,
            large_eval_budget=0,
            small_eval_budget=0,
            small_pop_active=True,
        )
        return WrapperState(strategy_state, restart_state)

    def ask(
        self,
        rng: chex.PRNGKey,
        state: WrapperState,
        params: Optional[WrapperParams] = None,
    ) -> Tuple[chex.Array, chex.ArrayTree]:
        """`ask` for new parameter candidates to evaluate next."""
        # Use default hyperparameters if no other settings provided
        if params is None:
            params = self.default_params
        # TODO: Cannot jit! Re-definition of strategy with different popsizes.
        # Is there a clever way to mask active members/popsize?
        # Only compile when base strategy is being updated with new popsize.
        rng_ask, rng_restart = jax.random.split(rng)
        if state.restart_state.restart_next:
            state = self.restart(rng_restart, state, params)
        x, strategy_state = self.base_strategy.ask(
            rng_ask, state.strategy_state, params.strategy_params
        )
        return x, state.replace(strategy_state=strategy_state)

    def restart(
        self,
        rng: chex.PRNGKey,
        state: chex.ArrayTree,
        params: chex.ArrayTree,
    ) -> chex.ArrayTree:
        """Reinstantiate a new strategy with interlaced population sizes."""
        # Track number of evals depending on active population
        large_eval_budget = jax.lax.select(
            state.restart_state.small_pop_active,
            state.restart_state.large_eval_budget,
            state.restart_state.large_eval_budget
            + state.restart_state.active_popsize
            * state.strategy_state.gen_counter,
        )
        small_eval_budget = jax.lax.select(
            state.restart_state.small_pop_active,
            state.restart_state.small_eval_budget
            + state.restart_state.active_popsize
            * state.strategy_state.gen_counter,
            state.restart_state.small_eval_budget,
        )
        small_pop_active = small_eval_budget < large_eval_budget

        # Update the population size based on active population size
        pop_mult = params.restart_params.popsize_multiplier ** (
            state.restart_state.restart_large_counter + 1
        )
        small_popsize = jax.lax.floor(
            self.default_popsize * pop_mult ** (jax.random.uniform(rng) ** 2)
        ).astype(int)
        large_popsize = self.default_popsize * pop_mult

        # Reinstantiate new strategy - based on name of previous strategy
        active_popsize = jax.lax.select(
            small_pop_active, small_popsize, large_popsize
        )

        # Reinstantiate new ES with new population size
        self.base_strategy = Strategies[self.base_strategy.strategy_name](
            popsize=int(active_popsize),
            num_dims=self.num_dims,
            **self.strategy_kwargs
        )

        strategy_state = self.base_strategy.initialize(
            rng, params.strategy_params
        )
        strategy_state = strategy_state.replace(
            mean=jax.lax.select(
                params.restart_params.copy_mean,
                state.strategy_state.mean,
                strategy_state.mean,
            ),
            best_fitness=state.strategy_state.best_fitness,
            best_member=state.strategy_state.best_member,
        )
        # Overwrite new state with old preservables
        restart_state = state.restart_state.replace(
            active_popsize=active_popsize,
            small_pop_active=small_pop_active,
            large_eval_budget=large_eval_budget,
            small_eval_budget=small_eval_budget,
            restart_large_counter=(
                state.restart_state.restart_large_counter + 1 - small_pop_active
            ),
            restart_counter=state.restart_state.restart_counter + 1,
            restart_next=False,
        )
        return WrapperState(strategy_state, restart_state)



from typing import Any
import functools
import sys
import chex
import jax
import jax.numpy as jnp
from flax import linen as nn
from ..core.fitness import (
    z_score_trafo,
    centered_rank_trafo,
    range_norm_trafo,
    compute_l2_norm,
)

if sys.version_info < (3, 8):
    # Load with pickle5 for python version compatibility
    import pickle5 as pickle
else:
    import pickle


def load_pkl_object(filename: Any, pkg_load: bool = False) -> chex.ArrayTree:
    """Reload pickle objects from path."""
    if not pkg_load:
        with open(filename, "rb") as input:
            obj = pickle.load(input)
    else:
        obj = pickle.loads(filename)
    return obj


def tanh_timestamp(x: chex.Array) -> chex.Array:
    """Timestamp embedding with evo-adapted timescales. (Metz et al., 2022)"""

    def single_frequency(timescale):
        return jnp.tanh(x / jnp.float32(timescale) - 1.0)

    all_frequencies = jnp.asarray(
        [1, 3, 10, 30, 50, 100, 250, 500, 750, 1000, 1250, 1500, 2000],
        dtype=jnp.float32,
    )
    return jax.vmap(single_frequency)(all_frequencies)


class EvolutionPath(object):
    def __init__(self, num_dims: int, timescales: chex.Array):
        self.num_dims = num_dims
        self.timescales = timescales

    def initialize(self) -> chex.Array:
        """Initialize evolution path arrays."""
        return jnp.zeros((self.num_dims, self.timescales.shape[0]))

    def update(self, paths: chex.Array, diff: chex.Array) -> chex.Array:
        """Batch update evolution paths for multiple dims & timescales."""

        def update_path(lrate, path, diff):
            return (1 - lrate) * path + lrate * diff

        return jax.vmap(update_path, in_axes=(0, 1, None), out_axes=1)(
            self.timescales, paths, diff
        )


class FitnessFeatures(object):
    """Fitness Feature Constructor."""

    def __init__(
        self,
        centered_rank: bool = False,
        z_score: bool = False,
        w_decay: float = 0.0,
        diff_best: bool = False,
        norm_range: bool = False,
        maximize: bool = False,
    ):
        self.centered_rank = centered_rank
        self.z_score = z_score
        self.w_decay = w_decay
        self.diff_best = diff_best
        self.norm_range = norm_range
        self.maximize = maximize

    @functools.partial(jax.jit, static_argnums=0)
    def apply(
        self, x: chex.Array, fitness: chex.Array, best_fitness: float
    ) -> chex.Array:
        """Compute and concatenate different fitness transformations."""
        fitness = jax.lax.select(self.maximize, -1 * fitness, fitness)
        fit_out = ((fitness < best_fitness) * 1.0).reshape(-1, 1)

        if self.centered_rank:
            fit_cr = centered_rank_trafo(fitness).reshape(-1, 1)
            fit_out = jnp.concatenate([fit_out, fit_cr], axis=1)
        if self.z_score:
            fit_zs = z_score_trafo(fitness).reshape(-1, 1)
            fit_out = jnp.concatenate([fit_out, fit_zs], axis=1)
        if self.diff_best:
            fit_best = norm_diff_best(fitness, best_fitness).reshape(-1, 1)
            fit_out = jnp.concatenate([fit_out, fit_best], axis=1)
        if self.norm_range:
            fit_norm = range_norm_trafo(fitness, -1.0, 1.0).reshape(-1, 1)
            fit_out = jnp.concatenate([fit_out, fit_norm], axis=1)
        if self.w_decay:
            fit_wnorm = compute_l2_norm(x).reshape(-1, 1)
            fit_out = jnp.concatenate([fit_out, fit_wnorm], axis=1)
        return fit_out


def norm_diff_best(fitness: chex.Array, best_fitness: float) -> chex.Array:
    """Normalizes difference from best previous fitness score."""
    fitness = jnp.clip(fitness, -1e10, 1e10)
    diff_best = fitness - best_fitness
    return jnp.clip(
        diff_best / (jnp.nanmax(diff_best) - jnp.nanmin(diff_best.min) + 1e-10),
        -1,
        1,
    )


class AttentionWeights(nn.Module):
    """Self-attention layer for recombination weights."""

    att_hidden_dims: int = 8

    @nn.compact
    def __call__(self, X: chex.Array) -> chex.Array:
        keys = nn.Dense(self.att_hidden_dims)(X)
        queries = nn.Dense(self.att_hidden_dims)(X)
        values = nn.Dense(1)(X)
        A = nn.softmax(jnp.matmul(queries, keys.T) / jnp.sqrt(X.shape[0]))
        weights = nn.softmax(jnp.matmul(A, values).squeeze())
        return weights[:, None]


class EvoPathMLP(nn.Module):
    """MLP layer for learning rate modulation based on evopaths."""

    mlp_hidden_dims: int = 8

    @nn.compact
    def __call__(
        self,
        path_c: chex.Array,
        path_sigma: chex.Array,
        time_embed: chex.Array,
    ):
        timestamps = jnp.repeat(
            jnp.expand_dims(time_embed, axis=0), repeats=path_c.shape[0], axis=0
        )
        X = jnp.concatenate([path_c, path_sigma, timestamps], axis=1)
        # Perform MLP hidden state update for each solution dim. in parallel
        hidden = jax.vmap(nn.Dense(self.mlp_hidden_dims), in_axes=(0))(X)
        hidden = nn.relu(hidden)
        lrates_mean = nn.sigmoid(nn.Dense(1)(hidden)).squeeze()
        lrates_sigma = nn.sigmoid(nn.Dense(1)(hidden)).squeeze()
        return lrates_mean, lrates_sigma



import jax
import jax.numpy as jnp
from evosax import OpenES, CMA_ES, IPOP_CMA_ES, BIPOP_CMA_ES
from evosax.restarts import Simple_Restarter, BIPOP_Restarter, IPOP_Restarter


def test_simple_restart():
    rng = jax.random.PRNGKey(0)
    strategy = CMA_ES(popsize=4, num_dims=2)
    re_strategy = Simple_Restarter(strategy)
    re_es_params = re_strategy.default_params
    re_es_params = re_es_params.replace(
        restart_params=re_es_params.restart_params.replace(min_num_gens=2)
    )
    state = re_strategy.initialize(rng, re_es_params)

    # Run 1st ask-tell-generation
    x, state = re_strategy.ask(rng, state, re_es_params)
    fitness = jnp.array([1.0, 2.0, 3.0, 4.0])
    state = re_strategy.tell(x, fitness, state, re_es_params)
    assert state.restart_state.restart_next == False

    # Run 2nd ask-tell-generation
    x, state = re_strategy.ask(rng, state, re_es_params)
    fitness = jnp.array([0.0, 0.0, 0.0, 0.0])
    state = re_strategy.tell(x, fitness, state, re_es_params)
    assert state.restart_state.restart_next == True

    x, state = re_strategy.ask(rng, state, re_es_params)
    assert (state.strategy_state.mean == jnp.zeros(2)).all()


def test_ipop_restart():
    rng = jax.random.PRNGKey(0)
    strategy = OpenES(popsize=4, num_dims=2)
    re_strategy = IPOP_Restarter(strategy)
    re_es_params = re_strategy.default_params
    re_es_params = re_es_params.replace(
        restart_params=re_es_params.restart_params.replace(min_num_gens=2)
    )
    state = re_strategy.initialize(rng, re_es_params)
    # Run 1st ask-tell-generation
    x, state = re_strategy.ask(rng, state, re_es_params)
    fitness = jnp.array([1.0, 2.0, 3.0, 4.0])
    state = re_strategy.tell(x, fitness, state, re_es_params)
    assert state.restart_state.restart_next == False
    assert state.restart_state.active_popsize == 4
    # Run 2nd ask-tell-generation
    x, state = re_strategy.ask(rng, state, re_es_params)
    fitness = jnp.array([0.0, 0.0, 0.0, 0.0])
    state = re_strategy.tell(x, fitness, state, re_es_params)
    assert state.restart_state.restart_next == True

    # Run 3rd ask-tell-generation
    x, state = re_strategy.ask(rng, state, re_es_params)
    assert (state.strategy_state.mean == jnp.zeros(2)).all()
    assert state.restart_state.active_popsize == 8

    fitness = jnp.array([1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0])
    state = re_strategy.tell(x, fitness, state, re_es_params)
    assert state.restart_state.restart_next == False
    assert state.restart_state.active_popsize == 8


def test_bipop_restart():
    rng = jax.random.PRNGKey(0)
    strategy = OpenES(popsize=4, num_dims=2)
    re_strategy = BIPOP_Restarter(strategy)
    re_es_params = re_strategy.default_params
    re_es_params = re_es_params.replace(
        restart_params=re_es_params.restart_params.replace(min_num_gens=2)
    )
    state = re_strategy.initialize(rng, re_es_params)
    # Run 1st ask-tell-generation
    x, state = re_strategy.ask(rng, state, re_es_params)
    fitness = jnp.array([1.0, 2.0, 3.0, 4.0])
    state = re_strategy.tell(x, fitness, state, re_es_params)
    assert state.restart_state.restart_next == False
    assert state.restart_state.active_popsize == 4

    # Run 2nd ask-tell-generation
    x, state = re_strategy.ask(rng, state, re_es_params)
    fitness = jnp.array([0.0, 0.0, 0.0, 0.0])
    state = re_strategy.tell(x, fitness, state, re_es_params)
    assert state.restart_state.restart_next == True

    # Run 3rd ask-tell-generation
    x, state = re_strategy.ask(rng, state, re_es_params)
    assert (state.strategy_state.mean == jnp.zeros(2)).all()
    assert state.restart_state.active_popsize == 8

    fitness = jnp.array([1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0])
    state = re_strategy.tell(x, fitness, state, re_es_params)
    assert state.restart_state.restart_next == False
    assert state.restart_state.active_popsize == 8

    # Run 4th ask-tell-generation
    x, state = re_strategy.ask(rng, state, re_es_params)
    fitness = jnp.array([0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0])
    state = re_strategy.tell(x, fitness, state, re_es_params)
    assert state.restart_state.restart_next == True

    x, state = re_strategy.ask(rng, state, re_es_params)
    assert (state.strategy_state.mean == jnp.zeros(2)).all()
    assert state.restart_state.active_popsize <= 8


def test_ipop_cma_es():
    rng = jax.random.PRNGKey(0)
    re_strategy = IPOP_CMA_ES(popsize=4, num_dims=2)
    re_es_params = re_strategy.default_params
    re_es_params = re_es_params.replace(
        restart_params=re_es_params.restart_params.replace(min_num_gens=2)
    )
    state = re_strategy.initialize(rng, re_es_params)
    # Run 1st ask-tell-generation
    x, state = re_strategy.ask(rng, state, re_es_params)
    fitness = jnp.array([1.0, 2.0, 3.0, 4.0])
    state = re_strategy.tell(x, fitness, state, re_es_params)
    assert state.restart_state.restart_next == False
    assert state.restart_state.active_popsize == 4
    # Run 2nd ask-tell-generation
    x, state = re_strategy.ask(rng, state, re_es_params)
    fitness = jnp.array([0.0, 0.0, 0.0, 0.0])
    state = re_strategy.tell(x, fitness, state, re_es_params)
    assert state.restart_state.restart_next == True
    # Run 3rd ask-tell-generation
    x, state = re_strategy.ask(rng, state, re_es_params)
    # assert (state.strategy_state.mean == jnp.zeros(2)).all()
    assert state.restart_state.active_popsize == 8

    fitness = jnp.array([1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0])
    state = re_strategy.tell(x, fitness, state, re_es_params)
    assert state.restart_state.restart_next == False
    assert state.restart_state.active_popsize == 8


def test_bipop_cma_es():
    rng = jax.random.PRNGKey(0)
    re_strategy = BIPOP_CMA_ES(popsize=4, num_dims=2)
    re_es_params = re_strategy.default_params
    re_es_params = re_es_params.replace(
        restart_params=re_es_params.restart_params.replace(min_num_gens=2)
    )
    state = re_strategy.initialize(rng, re_es_params)
    # Run 1st ask-tell-generation
    x, state = re_strategy.ask(rng, state, re_es_params)
    fitness = jnp.array([1.0, 2.0, 3.0, 4.0])
    state = re_strategy.tell(x, fitness, state, re_es_params)
    assert state.restart_state.restart_next == False
    assert state.restart_state.active_popsize == 4

    # Run 2nd ask-tell-generation
    x, state = re_strategy.ask(rng, state, re_es_params)
    fitness = jnp.array([0.0, 0.0, 0.0, 0.0])
    state = re_strategy.tell(x, fitness, state, re_es_params)
    assert state.restart_state.restart_next == True

    # Run 3rd ask-tell-generation
    x, state = re_strategy.ask(rng, state, re_es_params)
    # assert (state.strategy_state.mean == jnp.zeros(2)).all()
    assert state.restart_state.active_popsize == 8

    fitness = jnp.array([1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0])
    state = re_strategy.tell(x, fitness, state, re_es_params)
    assert state.restart_state.restart_next == False
    assert state.restart_state.active_popsize == 8

    # Run 4th ask-tell-generation
    x, state = re_strategy.ask(rng, state, re_es_params)
    fitness = jnp.array([0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0])
    state = re_strategy.tell(x, fitness, state, re_es_params)
    assert state.restart_state.restart_next == True

    x, state = re_strategy.ask(rng, state, re_es_params)
    # assert (state.strategy_state.mean == jnp.zeros(2)).all()
    assert state.restart_state.active_popsize <= 8



from typing import Optional, Union
import pickle
import jax
import jax.numpy as jnp
import chex
from functools import partial
from ..core.reshape import ParameterReshaper


class ESLog(object):
    def __init__(
        self,
        num_dims: Optional[int] = None,
        pholder_params: Optional[Union[chex.ArrayTree, chex.Array]] = None,
        num_generations: int = 200,
        top_k: int = 5,
        maximize: bool = False,
    ):
        """Simple jittable logging tool for ES rollouts."""
        # Setup optional parameter reshaper
        self.use_param_reshaper = pholder_params is not None
        if self.use_param_reshaper:
            self.param_reshaper = ParameterReshaper(pholder_params, n_devices=1)
            self.num_dims = self.param_reshaper.total_params
        else:
            self.num_dims = num_dims
        self.num_generations = num_generations
        self.top_k = top_k
        self.maximize = maximize

    @partial(jax.jit, static_argnums=(0,))
    def initialize(self) -> chex.ArrayTree:
        """Initialize the logger storage."""
        log = {
            "top_fitness": jnp.zeros(self.top_k)
            - 1e10 * self.maximize
            + 1e10 * (1 - self.maximize),
            "top_params": jnp.zeros((self.top_k, self.num_dims))
            - 1e10 * self.maximize
            + 1e10 * (1 - self.maximize),
            "log_top_1": jnp.zeros(self.num_generations)
            - 1e10 * self.maximize
            + 1e10 * (1 - self.maximize),
            "log_top_mean": jnp.zeros(self.num_generations)
            - 1e10 * self.maximize
            + 1e10 * (1 - self.maximize),
            "log_top_std": jnp.zeros(self.num_generations)
            - 1e10 * self.maximize
            + 1e10 * (1 - self.maximize),
            "log_gen_1": jnp.zeros(self.num_generations)
            - 1e10 * self.maximize
            + 1e10 * (1 - self.maximize),
            "log_gen_mean": jnp.zeros(self.num_generations)
            - 1e10 * self.maximize
            + 1e10 * (1 - self.maximize),
            "log_gen_std": jnp.zeros(self.num_generations)
            - 1e10 * self.maximize
            + 1e10 * (1 - self.maximize),
            "gen_counter": 0,
        }
        return log

    @partial(jax.jit, static_argnums=(0,))
    def update(
        self, log: chex.ArrayTree, x: chex.Array, fitness: chex.Array
    ) -> chex.ArrayTree:
        """Update the logging storage with newest data."""
        # Check if there are solutions better than current archive
        vals = jnp.hstack([log["top_fitness"], fitness])
        if self.use_param_reshaper:
            x = self.param_reshaper.flatten(x)
        params = jnp.vstack([log["top_params"], x])
        top_idx = self.maximize * ((-1) * vals).argsort() + (
            (1 - self.maximize) * vals.argsort()
        )
        log["top_fitness"] = vals[top_idx[: self.top_k]]
        log["top_params"] = params[top_idx[: self.top_k]]
        log["log_top_1"] = (
            log["log_top_1"].at[log["gen_counter"]].set(log["top_fitness"][0])
        )
        log["log_top_mean"] = (
            log["log_top_mean"]
            .at[log["gen_counter"]]
            .set(jnp.mean(log["top_fitness"]))
        )

        log["log_top_std"] = (
            log["log_top_std"]
            .at[log["gen_counter"]]
            .set(jnp.std(log["top_fitness"]))
        )
        log["log_gen_1"] = (
            log["log_gen_1"]
            .at[log["gen_counter"]]
            .set(
                self.maximize * jnp.max(fitness)
                + (1 - self.maximize) * jnp.min(fitness)
            )
        )
        log["log_gen_mean"] = (
            log["log_gen_mean"].at[log["gen_counter"]].set(jnp.mean(fitness))
        )
        log["log_gen_std"] = (
            log["log_gen_std"].at[log["gen_counter"]].set(jnp.std(fitness))
        )
        log["gen_counter"] += 1
        return log

    def save(self, log: chex.ArrayTree, filename: str):
        """Save different parts of logger in .pkl file."""
        with open(filename, "wb") as handle:
            pickle.dump(log, handle, protocol=pickle.HIGHEST_PROTOCOL)

    def load(self, filename: str):
        """Reload the pickle logger and return dictionary."""
        with open(filename, "rb") as handle:
            es_logger = pickle.load(handle)
        return es_logger

    def plot(
        self,
        log,
        title,
        ylims=None,
        fig=None,
        ax=None,
        no_legend=False,
    ):
        """Plot fitness trajectory from evo logger over generations."""
        import matplotlib.pyplot as plt

        if fig is None or ax is None:
            fig, ax = plt.subplots(1, 1, figsize=(6, 3))
        int_range = jnp.arange(1, log["gen_counter"] + 1)
        ax.plot(
            int_range, log["log_top_1"][: log["gen_counter"]], label="Top 1"
        )
        ax.plot(
            int_range,
            log["log_top_mean"][: log["gen_counter"]],
            label=f"Top-{self.top_k} Mean",
        )
        ax.plot(
            int_range, log["log_gen_1"][: log["gen_counter"]], label="Gen. 1"
        )
        ax.plot(
            int_range,
            log["log_gen_mean"][: log["gen_counter"]],
            label="Gen. Mean",
        )
        if ylims is not None:
            ax.set_ylim(ylims)
        if not no_legend:
            ax.legend()
        if title is not None:
            ax.set_title(title)
        ax.set_xlabel("Number of Generations")
        ax.set_ylabel("Fitness Score")
        ax.spines["top"].set_visible(False)
        ax.spines["right"].set_visible(False)
        return fig, ax



import jax
import jax.numpy as jnp
from evosax import Strategies
from evosax.problems import BBOBFitness
from evosax.core import FitnessShaper
from functools import partial

num_iters = 25


def test_strategy_run(strategy_name):
    # Loop over all strategies and test ask API
    rng = jax.random.PRNGKey(0)
    Strat = Strategies[strategy_name]
    # PBT also returns copy ID integer - treat separately
    if strategy_name == "ESMC":
        popsize = 21
    else:
        popsize = 20
    evaluator = BBOBFitness("Sphere", 2)
    fitness_shaper = FitnessShaper()

    batch_eval = evaluator.rollout
    strategy = Strat(popsize=popsize, num_dims=2)
    params = strategy.default_params
    state = strategy.initialize(rng, params)

    fitness_log = []
    for t in range(num_iters):
        rng, rng_eval, rng_iter = jax.random.split(rng, 3)
        x, state = strategy.ask(rng_iter, state, params)
        fitness = batch_eval(rng, x)
        fitness_shaped = fitness_shaper.apply(x, fitness)
        state = strategy.tell(x, fitness_shaped, state, params)
        best_id = jnp.argmin(fitness)
        fitness_log.append(fitness[best_id])
    # assert fitness[0] >= fitness[-1]


def test_strategy_scan(strategy_name):
    # Loop over all strategies and test ask API
    rng = jax.random.PRNGKey(0)
    Strat = Strategies[strategy_name]
    # PBT also returns copy ID integer - treat separately
    if strategy_name == "ESMC":
        popsize = 21
    else:
        popsize = 20
    evaluator = BBOBFitness("Sphere", 2)
    fitness_shaper = FitnessShaper()

    batch_eval = evaluator.rollout
    strategy = Strat(popsize=popsize, num_dims=2)
    es_params = strategy.default_params

    @partial(jax.jit, static_argnums=(1,))
    def run_plain_es(rng, num_steps):
        """Run evolution ask-eval-tell loop."""
        state = strategy.initialize(rng, es_params)

        def step(state_input, tmp):
            """Helper function to lax.scan through."""
            rng, state = state_input
            rng, rng_eval, rng_iter = jax.random.split(rng, 3)
            x, state = strategy.ask(rng_iter, state, es_params)
            fitness = batch_eval(rng_eval, x)
            fitness_shaped = fitness_shaper.apply(x, fitness)
            state = strategy.tell(x, fitness_shaped, state, es_params)
            best_id = jnp.argmin(fitness)
            return [rng, state], fitness[best_id]

        _, scan_out = jax.lax.scan(step, [rng, state], [jnp.zeros(num_steps)])
        return jnp.min(scan_out)

    run_plain_es(rng, num_iters)



from .control_gym import GymnaxFitness
from .vision import VisionFitness
from .bbob import BBOBFitness
from .sequence import SequenceFitness

ProblemMapper = {
    "Gymnax": GymnaxFitness,
    "Vision": VisionFitness,
    "BBOB": BBOBFitness,
    "Sequence": SequenceFitness,
}

__all__ = [
    "GymnaxFitness",
    "VisionFitness",
    "BBOBFitness",
    "SequenceFitness",
    "ProblemMapper",
]



from typing import Tuple, Optional, Union
import jax
import jax.numpy as jnp
import chex
from flax import struct
from ..strategy import Strategy
from ..core import GradientOptimizer, OptState, OptParams, exp_decay


@struct.dataclass
class EvoState:
    mean: chex.Array
    sigma: float
    pert_accum: chex.Array  # History of accum. noise perturb in partial unroll
    inner_step_counter: int  # Keep track of unner unroll steps for noise reset
    opt_state: OptState
    best_member: chex.Array
    best_fitness: float = jnp.finfo(jnp.float32).max
    gen_counter: int = 0


@struct.dataclass
class EvoParams:
    opt_params: OptParams
    T: int = 100  # Total inner problem length
    K: int = 10  # Truncation length for partial unrolls
    sigma_init: float = 0.1
    sigma_decay: float = 0.999
    sigma_limit: float = 0.1
    init_min: float = 0.0
    init_max: float = 0.0
    clip_min: float = -jnp.finfo(jnp.float32).max
    clip_max: float = jnp.finfo(jnp.float32).max


class PersistentES(Strategy):
    def __init__(
        self,
        popsize: int,
        num_dims: Optional[int] = None,
        pholder_params: Optional[Union[chex.ArrayTree, chex.Array]] = None,
        opt_name: str = "adam",
        lrate_init: float = 0.05,
        lrate_decay: float = 1.0,
        lrate_limit: float = 0.001,
        sigma_init: float = 0.03,
        sigma_decay: float = 1.0,
        sigma_limit: float = 0.01,
        mean_decay: float = 0.0,
        n_devices: Optional[int] = None,
        **fitness_kwargs: Union[bool, int, float]
    ):
        """Persistent ES (Vicol et al., 2021).
        Reference: http://proceedings.mlr.press/v139/vicol21a.html
        Inspired by: http://proceedings.mlr.press/v139/vicol21a/vicol21a-supp.pdf
        """
        super().__init__(
            popsize,
            num_dims,
            pholder_params,
            mean_decay,
            n_devices,
            **fitness_kwargs
        )
        assert not self.popsize & 1, "Population size must be even"
        assert opt_name in ["sgd", "adam", "rmsprop", "clipup", "adan"]
        self.optimizer = GradientOptimizer[opt_name](self.num_dims)
        self.strategy_name = "PersistentES"

        # Set core kwargs es_params (lrate/sigma schedules)
        self.lrate_init = lrate_init
        self.lrate_decay = lrate_decay
        self.lrate_limit = lrate_limit
        self.sigma_init = sigma_init
        self.sigma_decay = sigma_decay
        self.sigma_limit = sigma_limit

    @property
    def params_strategy(self) -> EvoParams:
        """Return default parameters of evolution strategy."""
        opt_params = self.optimizer.default_params.replace(
            lrate_init=self.lrate_init,
            lrate_decay=self.lrate_decay,
            lrate_limit=self.lrate_limit,
        )
        return EvoParams(
            opt_params=opt_params,
            sigma_init=self.sigma_init,
            sigma_decay=self.sigma_decay,
            sigma_limit=self.sigma_limit,
        )

    def initialize_strategy(
        self, rng: chex.PRNGKey, params: EvoParams
    ) -> chex.ArrayTree:
        """`initialize` the evolution strategy."""
        initialization = jax.random.uniform(
            rng,
            (self.num_dims,),
            minval=params.init_min,
            maxval=params.init_max,
        )
        state = EvoState(
            mean=initialization,
            pert_accum=jnp.zeros((self.popsize, self.num_dims)),
            opt_state=self.optimizer.initialize(params.opt_params),
            sigma=params.sigma_init,
            inner_step_counter=0,
            best_member=initialization,
        )
        return state

    def ask_strategy(
        self, rng: chex.PRNGKey, state: EvoState, params: EvoParams
    ) -> Tuple[chex.Array, EvoState]:
        """`ask` for new proposed candidates to evaluate next."""
        # Generate antithetic perturbations
        pos_perts = (
            jax.random.normal(rng, (self.popsize // 2, self.num_dims))
            * state.sigma
        )
        neg_perts = -pos_perts
        perts = jnp.concatenate([pos_perts, neg_perts], axis=0)
        # Add the perturbations from this unroll to the perturbation accumulators
        pert_accum = state.pert_accum + perts
        x = state.mean + perts
        return x, state.replace(pert_accum=pert_accum)

    def tell_strategy(
        self,
        x: chex.Array,
        fitness: chex.Array,
        state: EvoState,
        params: EvoParams,
    ) -> EvoState:
        """`tell` update to ES state."""
        theta_grad = jnp.mean(
            state.pert_accum * fitness.reshape(-1, 1) / (state.sigma ** 2),
            axis=0,
        )
        # Grad update using optimizer instance - decay lrate if desired
        mean, opt_state = self.optimizer.step(
            state.mean, theta_grad, state.opt_state, params.opt_params
        )
        opt_state = self.optimizer.update(opt_state, params.opt_params)
        inner_step_counter = state.inner_step_counter + params.K

        sigma = exp_decay(state.sigma, params.sigma_decay, params.sigma_limit)
        # Reset accumulated antithetic noise memory if done with inner problem
        reset = inner_step_counter >= params.T
        inner_step_counter = jax.lax.select(reset, 0, inner_step_counter)
        pert_accum = jax.lax.select(
            reset, jnp.zeros((self.popsize, self.num_dims)), state.pert_accum
        )
        return state.replace(
            mean=mean,
            sigma=sigma,
            opt_state=opt_state,
            pert_accum=pert_accum,
            inner_step_counter=inner_step_counter,
        )



from typing import Tuple, Optional, Union
import chex
from flax import struct
import jax
import jax.numpy as jnp
import pkgutil
from ..strategy import Strategy
from ..learned_eo.les_tools import load_pkl_object
from ..learned_eo.evotf_tools import (
    EvoTransformer,
    SolutionFeaturizer,
    FitnessFeaturizer,
    DistributionFeaturizer,
)


@struct.dataclass
class DistributionFeaturesState:
    old_mean: chex.Array
    old_sigma: chex.Array
    momentum_mean: chex.Array
    momentum_sigma: chex.Array
    evopath_mean: chex.Array
    evopath_sigma: chex.Array


@struct.dataclass
class FitnessFeaturesState:
    best_fitness: float


@struct.dataclass
class SolutionFeaturesState:
    best_fitness: float
    best_member: chex.Array
    gen_counter: int


@struct.dataclass
class EvoState:
    mean: chex.Array
    sigma: chex.Array
    sf_state: SolutionFeaturesState
    ff_state: FitnessFeaturesState
    df_state: DistributionFeaturesState
    solution_context: chex.Array
    fitness_context: chex.Array
    distribution_context: chex.Array
    best_member: chex.Array
    best_fitness: float = jnp.finfo(jnp.float32).max
    gen_counter: int = 0


@struct.dataclass
class EvoParams:
    net_params: chex.ArrayTree
    lrate_mean: float = 1.0
    lrate_sigma: float = 1.0
    sigma_init: float = 1.0
    init_min: float = 0.0
    init_max: float = 0.0
    clip_min: float = -jnp.finfo(jnp.float32).max
    clip_max: float = jnp.finfo(jnp.float32).max


class EvoTF_ES(Strategy):
    def __init__(
        self,
        popsize: int,
        num_dims: Optional[int] = None,
        pholder_params: Optional[Union[chex.ArrayTree, chex.Array]] = None,
        sigma_init: float = 1.0,
        max_context_len: int = 100,
        model_config: dict = dict(
            num_layers=1,
            num_heads=1,
            num_latents=8,
            latent_dim=32,
            embed_dim=128,
            use_fitness_encoder=True,
            use_dist_encoder=True,
            use_crossd_encoder=False,
        ),
        solution_config: dict = dict(
            norm_diff_mean=True,
            norm_diff_mean_sq=True,
            diff_best=True,
            norm_diff_best=True,
            maximize=False,
        ),
        fitness_config: dict = dict(
            improved_best=True,
            z_score=True,
            norm_diff_best=True,
            norm_range=True,
            snes_weights=True,
            des_weights=True,
            w_decay=0.00,
            maximize=False,
        ),
        distrib_config: dict = dict(
            use_mean=False,
            use_sigma=False,
            use_evopaths=True,
            use_momentum=True,
            evopath_timescales=[0.9, 0.95, 0.99],
            momentum_timescales=[0.9, 0.95, 0.99],
            use_oai_grad=True,
        ),
        use_antithetic_sampling: bool = False,
        net_params: Optional[chex.ArrayTree] = None,
        net_ckpt_path: Optional[str] = None,
        mean_decay: float = 0.0,
        n_devices: Optional[int] = None,
        **fitness_kwargs: Union[bool, int, float],
    ):
        super().__init__(
            popsize, num_dims, pholder_params, mean_decay, n_devices, **fitness_kwargs
        )
        self.strategy_name = "EvoTransformer"
        self.max_context_len = max_context_len
        self.model_config = model_config
        self.solution_config = solution_config
        self.fitness_config = fitness_config
        self.distrib_config = distrib_config
        self.model = EvoTransformer(**model_config)
        self.sf = SolutionFeaturizer(
            popsize=self.popsize,
            num_dims=self.num_dims,
            seq_len=self.max_context_len,
            **solution_config,
        )
        self.ff = FitnessFeaturizer(
            popsize=self.popsize,
            num_dims=self.num_dims,
            seq_len=self.max_context_len,
            **fitness_config,
        )
        self.df = DistributionFeaturizer(
            popsize=self.popsize,
            num_dims=self.num_dims,
            seq_len=self.max_context_len,
            **distrib_config,
        )
        if net_ckpt_path is not None:
            self.ckpt = load_pkl_object(net_ckpt_path)
            self.default_net_params = self.ckpt["net_params"]
            self.model_config = self.ckpt["model_config"]
            self.model = EvoTransformer(**self.model_config)
            self.sf = SolutionFeaturizer(
                popsize=self.popsize,
                num_dims=self.num_dims,
                seq_len=self.max_context_len,
                **self.ckpt["solution_config"],
            )
            self.ff = FitnessFeaturizer(
                popsize=self.popsize,
                num_dims=self.num_dims,
                seq_len=self.max_context_len,
                **self.ckpt["fitness_config"],
            )
            self.df = DistributionFeaturizer(
                popsize=self.popsize,
                num_dims=self.num_dims,
                seq_len=self.max_context_len,
                **self.ckpt["distrib_config"],
            )
            print(f"Loaded EvoTF model from ckpt: {net_ckpt_path}")

        if net_params is not None:
            self.default_net_params = net_params

        if net_params is None and net_ckpt_path is None:
            ckpt_fname = "2024_03_SNES_small.pkl"
            data = pkgutil.get_data(__name__, f"ckpt/evotf/{ckpt_fname}")
            self.ckpt = load_pkl_object(data, pkg_load=True)
            self.default_net_params = self.ckpt["net_params"]
            self.model_config = self.ckpt["model_config"]
            self.model = EvoTransformer(**self.model_config)
            self.sf = SolutionFeaturizer(
                popsize=self.popsize,
                num_dims=self.num_dims,
                seq_len=self.max_context_len,
                **self.ckpt["solution_config"],
            )
            self.ff = FitnessFeaturizer(
                popsize=self.popsize,
                num_dims=self.num_dims,
                seq_len=self.max_context_len,
                **self.ckpt["fitness_config"],
            )
            self.df = DistributionFeaturizer(
                popsize=self.popsize,
                num_dims=self.num_dims,
                seq_len=self.max_context_len,
                **self.ckpt["distrib_config"],
            )
            print(f"Loaded pretrained EvoTF model from ckpt: {ckpt_fname}")

        self.num_network_params = sum(
            x.size for x in jax.tree_leaves(self.default_net_params)
        )

        # Set core kwargs es_params
        self.sigma_init = sigma_init
        self.use_antithetic_sampling = use_antithetic_sampling
        if self.use_antithetic_sampling:
            assert not self.popsize & 1, "Population size must be even"
        # Setup look-ahead mask for forward pass
        self.la_mask = jnp.tril(jnp.ones((self.max_context_len, self.max_context_len)))
        self.strategy_name = "EvoTF_ES"

    @property
    def params_strategy(self) -> EvoParams:
        """Return default parameters of evolutionary strategy."""
        params = EvoParams(
            sigma_init=self.sigma_init,
            net_params=self.default_net_params,
        )
        return params

    def initialize_strategy(self, rng: chex.PRNGKey, params: EvoParams) -> EvoState:
        """`initialize` the evolutionary strategy."""
        initialization = jax.random.uniform(
            rng,
            (self.num_dims,),
            minval=params.init_min,
            maxval=params.init_max,
        )
        scon_shape = (
            1,
            self.max_context_len,
            self.popsize,
            self.num_dims,
            self.sf.num_features,
        )
        fcon_shape = (
            1,
            self.max_context_len,
            self.popsize,
            self.ff.num_features,
        )
        dcon_shape = (
            1,
            self.max_context_len,
            self.num_dims,
            self.df.num_features,
        )
        state = EvoState(
            mean=initialization,
            sigma=params.sigma_init * jnp.ones(self.num_dims),
            sf_state=self.sf.initialize(),
            ff_state=self.ff.initialize(),
            df_state=self.df.initialize(),
            solution_context=jnp.zeros(scon_shape),
            fitness_context=jnp.zeros(fcon_shape),
            distribution_context=jnp.zeros(dcon_shape),
            best_member=initialization,
        )

        return state

    def ask_strategy(
        self, rng: chex.PRNGKey, state: EvoState, params: EvoParams
    ) -> Tuple[chex.Array, EvoState]:
        """`ask` for new parameter candidates to evaluate next."""
        if not self.use_antithetic_sampling:
            noise = jax.random.normal(rng, (self.popsize, self.num_dims))
        else:
            noise_p = jax.random.normal(rng, (int(self.popsize / 2), self.num_dims))
            noise = jnp.concatenate([noise_p, -noise_p], axis=0)
        x = state.mean + noise * state.sigma.reshape(1, self.num_dims)
        return x, state

    def tell_strategy(
        self,
        x: chex.Array,
        fitness: chex.Array,
        state: EvoState,
        params: EvoParams,
    ) -> EvoState:
        """`tell` performance data for strategy state update."""
        # Get features from population info
        sfeatures, sf_state = self.sf.featurize(
            x, fitness, state.mean, state.sigma, state.sf_state
        )
        ffeatures, ff_state = self.ff.featurize(x, fitness, state.ff_state)
        dfeatures, df_state = self.df.featurize(
            x, fitness, state.mean, state.sigma, state.df_state
        )

        # Update the context with a sliding window
        shift_buffer = state.gen_counter >= self.max_context_len
        idx = jax.lax.select(shift_buffer, self.max_context_len - 1, state.gen_counter)
        solution_context = jax.lax.select(
            shift_buffer,
            state.solution_context.at[0, :-1].set(state.solution_context[0, 1:]),
            state.solution_context,
        )
        fitness_context = jax.lax.select(
            shift_buffer,
            state.fitness_context.at[0, :-1].set(state.fitness_context[0, 1:]),
            state.fitness_context,
        )
        distribution_context = jax.lax.select(
            shift_buffer,
            state.distribution_context.at[0, :-1].set(
                state.distribution_context[0, 1:]
            ),
            state.distribution_context,
        )
        solution_context = solution_context.at[0, idx].set(sfeatures)
        fitness_context = fitness_context.at[0, idx].set(ffeatures)
        distribution_context = distribution_context.at[0, idx].set(dfeatures)

        # Update strategy via evotf forward pass
        @jax.jit
        def infer_step(solution_context, fitness_context, distribution_context):
            return self.model.apply(
                params.net_params,
                solution_context,
                fitness_context,
                distribution_context,
                rngs={"dropout": jax.random.PRNGKey(0)},
                train=False,
                mask=self.la_mask,
            )

        pred, att = infer_step(solution_context, fitness_context, distribution_context)
        pred_mean = state.mean + params.lrate_mean * state.sigma * pred[0, 0, idx]
        pred_sigma = state.sigma * jnp.exp(params.lrate_sigma / 2 * pred[1, 0, idx])
        pred_sigma = jnp.clip(pred_sigma, 1e-08)
        # Collect and sort attention by population fitness order
        return state.replace(
            mean=pred_mean,
            sigma=pred_sigma,
            sf_state=sf_state,
            ff_state=ff_state,
            df_state=df_state,
            solution_context=solution_context,
            fitness_context=fitness_context,
            distribution_context=distribution_context,
        )



from typing import Tuple, Optional, Union
import jax
import jax.numpy as jnp
import chex
from flax import struct
from .snes import get_snes_weights
from ..strategy import Strategy


@struct.dataclass
class EvoState:
    mean: chex.Array
    sigma: float
    B: chex.Array
    noise: chex.Array
    lrate_sigma: float
    weights: chex.Array
    best_member: chex.Array
    best_fitness: float = jnp.finfo(jnp.float32).max
    gen_counter: int = 0


@struct.dataclass
class EvoParams:
    lrate_mean: float = 1.0
    lrate_sigma_init: float = 0.1
    lrate_B: float = 0.1
    sigma_init: float = 1.0
    use_adasam: bool = False  # Adaptation sampling lrate sigma
    rho: float = 0.5  # Significance level adaptation sampling
    c_prime: float = 0.1  # Adaptation sampling step size
    init_min: float = 0.0
    init_max: float = 0.0
    clip_min: float = -jnp.finfo(jnp.float32).max
    clip_max: float = jnp.finfo(jnp.float32).max


class xNES(Strategy):
    def __init__(
        self,
        popsize: int,
        num_dims: Optional[int] = None,
        pholder_params: Optional[Union[chex.ArrayTree, chex.Array]] = None,
        sigma_init: float = 1.0,
        mean_decay: float = 0.0,
        n_devices: Optional[int] = None,
        **fitness_kwargs: Union[bool, int, float]
    ):
        """Exponential Natural ES (Wierstra et al., 2014)
        Reference: https://www.jmlr.org/papers/volume15/wierstra14a/wierstra14a.pdf
        Inspired by: https://github.com/chanshing/xnes"""
        super().__init__(
            popsize, num_dims, pholder_params, mean_decay, n_devices, **fitness_kwargs
        )
        self.strategy_name = "xNES"

        # Set core kwargs es_params
        self.sigma_init = sigma_init

    @property
    def params_strategy(self) -> EvoParams:
        """Return default parameters of evolutionary strategy."""
        lrate_sigma = (9 + 3 * jnp.log(self.num_dims)) / (
            5 * jnp.sqrt(self.num_dims) * self.num_dims
        )
        rho = 0.5 - 1.0 / (3 * (self.num_dims + 1))
        params = EvoParams(
            lrate_sigma_init=lrate_sigma,
            lrate_B=lrate_sigma,
            rho=rho,
            sigma_init=self.sigma_init,
        )
        return params

    def initialize_strategy(self, rng: chex.PRNGKey, params: EvoParams) -> EvoState:
        """`initialize` the evolutionary strategy."""
        initialization = jax.random.uniform(
            rng,
            (self.num_dims,),
            minval=params.init_min,
            maxval=params.init_max,
        )
        weights = get_snes_weights(self.popsize)
        state = EvoState(
            mean=initialization,
            B=jnp.eye(self.num_dims) * params.sigma_init,
            sigma=params.sigma_init,
            noise=jnp.zeros((self.popsize, self.num_dims)),
            lrate_sigma=params.lrate_sigma_init,
            weights=weights,
            best_member=initialization,
        )

        return state

    def ask_strategy(
        self, rng: chex.PRNGKey, state: EvoState, params: EvoParams
    ) -> Tuple[chex.Array, EvoState]:
        """`ask` for new parameter candidates to evaluate next."""
        noise = jax.random.normal(rng, (self.popsize, self.num_dims))

        def scale_orient(n, sigma, B):
            return sigma * B.T @ n

        scaled_noise = jax.vmap(scale_orient, in_axes=(0, None, None))(
            noise, state.sigma, state.B
        )
        x = state.mean + scaled_noise
        return x, state.replace(noise=noise)

    def tell_strategy(
        self,
        x: chex.Array,
        fitness: chex.Array,
        state: EvoState,
        params: EvoParams,
    ) -> EvoState:
        """`tell` performance data for strategy state update."""
        ranks = fitness.argsort()
        sorted_noise = state.noise[ranks]
        grad_mean = (state.weights * sorted_noise).sum(axis=0)

        def s_grad_m(weight, noise):
            return weight * (noise @ noise.T - jnp.eye(self.num_dims))

        grad_m = jax.vmap(s_grad_m, in_axes=(0, 0))(state.weights, sorted_noise).sum(
            axis=0
        )
        grad_sigma = jnp.trace(grad_m) / self.num_dims
        grad_B = grad_m - grad_sigma * jnp.eye(self.num_dims)

        mean = state.mean + params.lrate_mean * state.sigma * state.B @ grad_mean
        sigma = state.sigma * jnp.exp(state.lrate_sigma / 2 * grad_sigma)
        B = state.B * jnp.exp(params.lrate_B / 2 * grad_B)

        lrate_sigma = adaptation_sampling(
            state.lrate_sigma,
            params.lrate_sigma_init,
            mean,
            B,
            sigma,
            state.sigma,
            sorted_noise,
            params.c_prime,
            params.rho,
        )
        lrate_sigma = jax.lax.select(params.use_adasam, lrate_sigma, state.lrate_sigma)
        return state.replace(mean=mean, sigma=sigma, B=B, lrate_sigma=lrate_sigma)


def adaptation_sampling(
    lrate_sigma: float,
    lrate_sigma_init: float,
    mean: chex.Array,
    B: chex.Array,
    sigma: float,
    sigma_old: float,
    sorted_noise: chex.Array,
    c_prime: float,
    rho: float,
) -> float:
    """Adaptation sampling on sigma/std learning rate."""
    BB = B.T @ B
    A = sigma**2 * BB
    sigma_prime = sigma * jnp.sqrt(sigma / sigma_old)
    A_prime = sigma_prime**2 * BB

    # Probability ration and u-test - sorted order assumed for noise
    prob_0 = jax.scipy.stats.multivariate_normal.logpdf(sorted_noise, mean, A)
    prob_1 = jax.scipy.stats.multivariate_normal.logpdf(sorted_noise, mean, A_prime)
    w = jnp.exp(prob_1 - prob_0)
    popsize = sorted_noise.shape[0]
    n = jnp.sum(w)
    u = jnp.sum(w * (jnp.arange(popsize) + 0.5))
    u_mean = popsize * n / 2
    u_sigma = jnp.sqrt(popsize * n * (popsize + n + 1) / 12)
    cumulative = jax.scipy.stats.norm.cdf(u, loc=u_mean + 1e-10, scale=u_sigma + 1e-10)

    # Check test significance and update lrate
    lrate_sigma = jax.lax.select(
        cumulative < rho,
        (1 - c_prime) * lrate_sigma + c_prime * lrate_sigma_init,
        jnp.minimum(1, (1 - c_prime) * lrate_sigma),
    )
    return lrate_sigma



import chex
import jax.numpy as jnp
from ..utils.eigen_decomp import full_eigen_decomp


def min_gen_criterion(
    fitness: chex.Array, state: chex.ArrayTree, params: chex.ArrayTree
) -> bool:
    """Allow stopping if minimal number of generations has passed."""
    min_gen_passed = (
        state.strategy_state.gen_counter >= params.restart_params.min_num_gens
    )
    return min_gen_passed


def spread_criterion(
    fitness: chex.Array, state: chex.ArrayTree, params: chex.ArrayTree
) -> bool:
    """Stop if min/max fitness spread of recent generation is below thresh."""
    fit_var_too_low = (
        jnp.max(fitness) - jnp.min(fitness)
        < params.restart_params.min_fitness_spread
    )
    return fit_var_too_low


def cma_criterion(
    fitness: chex.Array, state: chex.ArrayTree, params: chex.ArrayTree
) -> bool:
    """Termination criterion specific to CMA-ES strategy. Default tolerances:
    tol_x - 1e-12 * sigma
    tol_x_up - 1e4
    tol_condition_C - 1e14
    """
    cma_term = 0
    dC = jnp.diag(state.strategy_state.C)
    # Note: Criterion requires full covariance matrix for decomposition!
    C, B, D = full_eigen_decomp(
        state.strategy_state.C,
        state.strategy_state.B,
        state.strategy_state.D,
        state.strategy_state.gen_counter,
    )

    # Stop if std of normal distrib is smaller than tolx in all coordinates
    # and pc is smaller than tolx in all components.
    cond_s_1 = jnp.all(
        state.strategy_state.sigma * dC < params.restart_params.tol_x
    )
    cond_s_2 = jnp.all(
        state.strategy_state.sigma * state.strategy_state.p_c
        < params.restart_params.tol_x
    )
    cma_term += jnp.logical_and(cond_s_1, cond_s_2)

    # Stop if detecting divergent behavior -> Stepsize sigma exploded.
    cma_term += (
        state.strategy_state.sigma * jnp.max(D) > params.restart_params.tol_x_up
    )

    # No effect coordinates: stop if adding 0.2-standard deviations
    # in any single coordinate does not change mean.
    cond_no_coord_change = jnp.any(
        state.strategy_state.mean
        == state.strategy_state.mean
        + (0.2 * state.strategy_state.sigma * jnp.sqrt(dC))
    )
    cma_term += cond_no_coord_change

    # No effect axis: stop if adding 0.1-standard deviation vector in
    # any principal axis direction of C does not change m.
    cond_no_axis_change = jnp.all(
        state.strategy_state.mean
        == state.strategy_state.mean
        + (0.1 * state.strategy_state.sigma * D[0] * B[:, 0])
    )
    cma_term += cond_no_axis_change

    # Stop if the condition number of the covariance matrix exceeds 1e14.
    cond_condition_cov = (
        jnp.max(D) / jnp.min(D) > params.restart_params.tol_condition_C
    )
    cma_term += cond_condition_cov
    return cma_term > 0


def amalgam_criterion(
    fitness: chex.Array, state: chex.ArrayTree, params: chex.ArrayTree
) -> bool:
    """Termination criterion for iAMaLGaM algorithm (Bosman et al. 2013)"""
    return state.strategy_state.c_mult < 1e-10



from typing import Tuple, Optional, List
from flax import linen as nn
import jax.numpy as jnp
import chex
from .shared import scaled_dot_product, expand_mask, MLP, PositionalEncoding


class MultiheadAttention(nn.Module):
    embed_dim: int
    num_heads: int
    dropout_prob: float = 0.0
    use_bias: bool = False
    out_att_maps: bool = False

    def setup(self):
        self.qkv_proj = nn.Dense(
            features=3 * self.embed_dim,
            kernel_init=nn.initializers.xavier_uniform(),
            bias_init=nn.initializers.zeros,
            use_bias=self.use_bias,
        )
        self.out_proj = nn.Dense(
            features=self.embed_dim,
            kernel_init=nn.initializers.xavier_uniform(),
            bias_init=nn.initializers.zeros,
            use_bias=self.use_bias,
        )
        self.attn_dropout = nn.Dropout(self.dropout_prob)
        self.resid_dropout = nn.Dropout(self.dropout_prob)

    def __call__(
        self,
        x: chex.Array,
        mask: Optional[chex.Array] = None,
        train: bool = True,
    ) -> Tuple[chex.Array, chex.Array]:
        batch_size, seq_length, embed_dim = x.shape
        if mask is not None:
            mask = expand_mask(mask)
        qkv = self.qkv_proj(x)
        qkv = qkv.reshape(batch_size, seq_length, self.num_heads, -1)
        qkv = qkv.transpose(0, 2, 1, 3)
        q, k, v = jnp.array_split(qkv, 3, axis=-1)

        attention = scaled_dot_product(q, k, mask)
        attention = self.attn_dropout(attention, deterministic=not train)
        values = jnp.matmul(attention, v)
        values = values.transpose(0, 2, 1, 3)
        values = values.reshape(batch_size, seq_length, embed_dim)
        out = self.out_proj(values)
        out = self.resid_dropout(out, deterministic=not train)
        if self.out_att_maps:
            return out, attention
        else:
            return out, None


class AttentionBlock(nn.Module):
    num_heads: int
    embed_dim: int
    dropout_prob: float
    use_bias: bool
    out_att_maps: bool

    def setup(self):
        self.ln_1 = nn.LayerNorm(use_bias=self.use_bias)
        self.attn = MultiheadAttention(
            self.embed_dim,
            self.num_heads,
            self.dropout_prob,
            self.use_bias,
            self.out_att_maps,
        )
        self.ln_2 = nn.LayerNorm(use_bias=self.use_bias)
        self.mlp = MLP(self.embed_dim, self.dropout_prob, self.use_bias)

    def __call__(
        self,
        x: chex.Array,
        mask: Optional[chex.Array] = None,
        train: bool = True,
    ) -> Tuple[chex.Array, chex.Array]:
        attn_out, attn = self.attn(self.ln_1(x), mask, train)
        x = x + attn_out
        x = x + self.mlp(self.ln_2(x), train)
        return x, attn


class AttentionEncoder(nn.Module):
    embed_dim: int
    num_heads: int
    num_layers: int
    dropout_prob: float = 0.0
    input_dropout_prob: float = 0.0
    use_bias: bool = False
    out_att_maps: bool = False

    def setup(self):
        self.input_dropout = nn.Dropout(self.input_dropout_prob)
        self.input_layer = nn.Dense(self.embed_dim, use_bias=self.use_bias)
        self.positional_encoding = PositionalEncoding(self.embed_dim)
        self.transformer = [
            AttentionBlock(
                num_heads=self.num_heads,
                embed_dim=self.embed_dim,
                dropout_prob=self.dropout_prob,
                use_bias=self.use_bias,
                out_att_maps=self.out_att_maps,
            )
            for _ in range(self.num_layers)
        ]

    def __call__(
        self,
        x: chex.Array,
        mask: Optional[chex.Array] = None,
        add_positional_encoding: bool = True,
        train=True,
    ) -> Tuple[chex.Array, List[chex.Array]]:
        x = self.input_layer(x)
        if add_positional_encoding:
            x = self.positional_encoding(x)
        x = self.input_dropout(x, deterministic=not train)
        # Loop over transformer blocks and collect attention maps
        attn_maps = []
        for layer in self.transformer:
            x, attn = layer(x, mask, train)
            attn_maps.append(attn)
        return x, attn_maps



import chex
import jax
import jax.numpy as jnp
from evojax.algo.base import NEAlgorithm
from evosax import Strategy


class Evosax2JAX_Wrapper(NEAlgorithm):
    """Wrapper for evosax-style ES for EvoJAX deployment."""

    def __init__(
        self,
        evosax_strategy: Strategy,
        param_size: int,
        pop_size: int,
        es_config: dict = {},
        es_params: dict = {},
        opt_params: dict = {},
        seed: int = 42,
    ):
        self.es = evosax_strategy(
            popsize=pop_size, num_dims=param_size, **es_config, **opt_params
        )
        self.es_params = self.es.default_params.replace(**es_params)
        self.pop_size = pop_size
        self.param_size = param_size
        self.rand_key = jax.random.PRNGKey(seed=seed)
        self.rand_key, init_key = jax.random.split(self.rand_key)
        self.es_state = self.es.initialize(init_key, self.es_params)

    def ask(self) -> chex.Array:
        """Ask strategy for next set of solution candidates to evaluate."""
        self.rand_key, ask_key = jax.random.split(self.rand_key)
        self.params, self.es_state = self.es.ask(
            ask_key, self.es_state, self.es_params
        )
        return self.params

    def tell(self, fitness: chex.Array) -> None:
        """Tell strategy about most recent fitness evaluations."""
        self.es_state = self.es.tell(
            self.params, fitness, self.es_state, self.es_params
        )

    @property
    def best_params(self) -> chex.Array:
        """Return set of mean/best parameters."""
        return jnp.array(self.es_state.mean, copy=True)

    @best_params.setter
    def best_params(self, params: chex.Array) -> None:
        """Update the best parameters stored internally."""
        self.es_state = self.es_state.replace(mean=jnp.array(params, copy=True))

    @property
    def solution(self):
        """Get evaluation parameters for current ES state."""
        return self.es.get_eval_params(self.es_state)



import functools
from typing import Tuple
import chex
from flax import struct
import jax
import jax.numpy as jnp
from evosax.core.fitness import range_norm_trafo
from evosax.utils.helpers import get_best_fitness_member


@struct.dataclass
class SolutionFeaturesState:
    best_fitness: float
    best_member: chex.Array
    gen_counter: int


class SolutionFeaturizer(object):
    def __init__(
        self,
        popsize: int,
        num_dims: int,
        seq_len: int,
        norm_diff_mean: bool = False,
        norm_diff_mean_sq: bool = False,
        diff_best: bool = False,
        norm_diff_best: bool = False,
        maximize: bool = False,
        verbose: bool = False,
    ):
        self.popsize = popsize
        self.num_dims = num_dims
        self.seq_len = seq_len
        self.norm_diff_mean = norm_diff_mean
        self.norm_diff_mean_sq = norm_diff_mean_sq
        self.diff_best = diff_best
        self.norm_diff_best = norm_diff_best
        self.maximize = maximize
        self.verbose = verbose

        if self.verbose:
            print(
                f"Solution Features / Batch shape: {self.num_features} / {self.example_batch_shape}"
            )
            print("[BASE] Diff mean -> (x - mu)")
            print(f"[{self.norm_diff_mean}] Norm difference mean -> (x - mu)/ sigma")
            print(
                f"[{self.norm_diff_mean_sq}] Norm difference to mean squared -> [(x - mu)/ sigma]^2"
            )
            print(f"[{self.diff_best}] Difference to best -> (x - x_best)")
            print(
                f"[{self.norm_diff_best}] Normalized difference to best  -> (x - x_best) in [-0.5, 0.5]"
            )

    @functools.partial(jax.jit, static_argnums=0)
    def initialize(self):
        return SolutionFeaturesState(
            best_fitness=jnp.finfo(jnp.float32).max,
            best_member=jnp.zeros((self.num_dims,)),
            gen_counter=0,
        )

    @functools.partial(jax.jit, static_argnums=0)
    def featurize(
        self,
        x: chex.Array,
        fitness: chex.Array,
        mean: chex.Array,
        sigma: chex.Array,
        state: SolutionFeaturesState,
    ) -> chex.Array:
        diff_mean = x - mean
        sol_out = jnp.expand_dims(x, axis=-1)

        if self.norm_diff_mean:
            norm_diff_mean = jnp.expand_dims(diff_mean / sigma, axis=-1)
            sol_out = jnp.concatenate([sol_out, norm_diff_mean], axis=-1)

        if self.norm_diff_mean_sq:
            norm_diff_mean_sq = jnp.expand_dims(jnp.square(diff_mean / sigma), axis=-1)
            sol_out = jnp.concatenate([sol_out, norm_diff_mean_sq], axis=-1)

        best_member, best_fitness = get_best_fitness_member(
            x, fitness, state, self.maximize
        )
        if self.diff_best:
            diff_best = jnp.expand_dims(x - best_member, axis=-1)
            sol_out = jnp.concatenate([sol_out, diff_best], axis=-1)

        if self.norm_diff_best:
            dist_best_norm = jax.vmap(
                range_norm_trafo, in_axes=(1, None, None), out_axes=1
            )(x - best_member, -0.5, 0.5)
            dist_best_norm = jnp.expand_dims(dist_best_norm, axis=-1)
            sol_out = jnp.concatenate([sol_out, dist_best_norm], axis=-1)
        return sol_out, state.replace(
            best_member=best_member,
            best_fitness=best_fitness,
            gen_counter=state.gen_counter + 1,
        )

    @property
    def num_features(self) -> int:
        return (
            1
            + self.norm_diff_mean
            + self.norm_diff_mean_sq
            + self.diff_best
            + self.norm_diff_best
        )

    @property
    def example_batch_shape(self) -> Tuple[int, ...]:
        return (
            1,  # Batchsize
            self.seq_len,  # Timesteps
            self.popsize,  # Popsize
            self.num_dims,  # Number of dims
            self.num_features,  # Solution features
        )



import jax
import jax.numpy as jnp
import chex
from typing import Union, Optional
from jax import vjp, flatten_util
from jax.tree_util import tree_flatten


def ravel_pytree(pytree):
    leaves, _ = tree_flatten(pytree)
    flat, _ = vjp(ravel_list, *leaves)
    return flat


def ravel_list(*lst):
    return (
        jnp.concatenate([jnp.ravel(elt) for elt in lst])
        if lst
        else jnp.array([])
    )


class ParameterReshaper(object):
    def __init__(
        self,
        placeholder_params: Union[chex.ArrayTree, chex.Array],
        n_devices: Optional[int] = None,
        verbose: bool = True,
    ):
        """Reshape flat parameters vectors into generation eval shape."""
        # Get network shape to reshape
        self.placeholder_params = placeholder_params

        # Set total parameters depending on type of placeholder params
        flat, self.unravel_pytree = flatten_util.ravel_pytree(
            placeholder_params
        )
        self.total_params = flat.shape[0]
        self.reshape_single = jax.jit(self.unravel_pytree)

        if n_devices is None:
            self.n_devices = jax.local_device_count()
        else:
            self.n_devices = n_devices
        if self.n_devices > 1 and verbose:
            print(
                f"ParameterReshaper: {self.n_devices} devices detected. Please"
                " make sure that the ES population size divides evenly across"
                " the number of devices to pmap/parallelize over."
            )

        if verbose:
            print(
                f"ParameterReshaper: {self.total_params} parameters detected"
                " for optimization."
            )

    def reshape(self, x: chex.Array) -> chex.ArrayTree:
        """Perform reshaping for a 2D matrix (pop_members, params)."""
        vmap_shape = jax.vmap(self.reshape_single)
        if self.n_devices > 1:
            x = self.split_params_for_pmap(x)
            map_shape = jax.pmap(vmap_shape)
        else:
            map_shape = vmap_shape
        return map_shape(x)

    def multi_reshape(self, x: chex.Array) -> chex.ArrayTree:
        """Reshape parameters lying already on different devices."""
        # No reshaping required!
        vmap_shape = jax.vmap(self.reshape_single)
        return jax.pmap(vmap_shape)(x)

    def flatten(self, x: chex.ArrayTree) -> chex.Array:
        """Reshaping pytree parameters (population) into flat array."""
        vmap_flat = jax.vmap(ravel_pytree)
        if self.n_devices > 1:
            # Flattening of pmap paramater trees to apply vmap flattening
            def map_flat(x):
                x_re = jax.tree_map(lambda x: x.reshape(-1, *x.shape[2:]), x)
                return vmap_flat(x_re)

        else:
            map_flat = vmap_flat
        flat = map_flat(x)
        # Out shape: (pop, params)
        return flat

    def flatten_single(self, x: chex.ArrayTree) -> chex.Array:
        """Reshaping pytree parameters (single) into flat array."""
        return ravel_pytree(x)

    def multi_flatten(self, x: chex.Array) -> chex.ArrayTree:
        """Flatten parameters lying remaining on different devices."""
        # No reshaping required!
        vmap_flat = jax.vmap(ravel_pytree)
        return jax.pmap(vmap_flat)(x)

    def split_params_for_pmap(self, param: chex.Array) -> chex.Array:
        """Helper reshapes param (bs, #params) into (#dev, bs/#dev, #params)."""
        return jnp.stack(jnp.split(param, self.n_devices))

    @property
    def vmap_dict(self) -> chex.ArrayTree:
        """Get a dictionary specifying axes to vmap over."""
        vmap_dict = jax.tree_map(lambda x: 0, self.placeholder_params)
        return vmap_dict



from typing import Tuple, Optional, Union
import jax
import jax.numpy as jnp
import chex
from flax import struct
from ..strategy import Strategy


@struct.dataclass
class EvoState:
    mean: chex.Array
    archive: chex.Array
    fitness: chex.Array
    sigma: chex.Array
    best_member: chex.Array
    best_fitness: float = jnp.finfo(jnp.float32).max
    gen_counter: int = 0


@struct.dataclass
class EvoParams:
    sigma_init: float = 0.07
    sigma_meta: float = 2.0
    sigma_best_limit: float = 0.0001
    init_min: float = 0.0
    init_max: float = 0.0
    clip_min: float = -jnp.finfo(jnp.float32).max
    clip_max: float = jnp.finfo(jnp.float32).max


class SAMR_GA(Strategy):
    def __init__(
        self,
        popsize: int,
        num_dims: Optional[int] = None,
        pholder_params: Optional[Union[chex.ArrayTree, chex.Array]] = None,
        elite_ratio: float = 0.0,
        sigma_init: float = 0.07,
        sigma_meta: float = 2.0,
        n_devices: Optional[int] = None,
        **fitness_kwargs: Union[bool, int, float]
    ):
        """Self-Adaptation Mutation Rate (SAMR) GA."""

        super().__init__(
            popsize,
            num_dims,
            pholder_params,
            n_devices=n_devices,
            **fitness_kwargs
        )
        self.elite_ratio = elite_ratio
        self.elite_popsize = max(1, int(self.popsize * self.elite_ratio))
        self.strategy_name = "SAMR_GA"

        # Set core kwargs es_params
        self.sigma_init = sigma_init
        self.sigma_meta = sigma_meta

    @property
    def params_strategy(self) -> EvoParams:
        """Return default parameters of evolution strategy."""
        return EvoParams(sigma_init=self.sigma_init, sigma_meta=self.sigma_meta)

    def initialize_strategy(
        self, rng: chex.PRNGKey, params: EvoParams
    ) -> EvoState:
        """`initialize` the differential evolution strategy."""
        initialization = jax.random.uniform(
            rng,
            (self.elite_popsize, self.num_dims),
            minval=params.init_min,
            maxval=params.init_max,
        )
        state = EvoState(
            mean=initialization.mean(axis=0),
            archive=initialization,
            fitness=jnp.zeros(self.elite_popsize) + jnp.finfo(jnp.float32).max,
            sigma=jnp.zeros(self.elite_popsize) + params.sigma_init,
            best_member=initialization.mean(axis=0),
        )
        return state

    def ask_strategy(
        self, rng: chex.PRNGKey, state: EvoState, params: EvoParams
    ) -> Tuple[chex.Array, EvoState]:
        """`ask` for new proposed candidates to evaluate next."""
        rng, rng_idx, rng_eps_x, rng_eps_s = jax.random.split(rng, 4)
        eps_x = jax.random.normal(rng_eps_x, (self.popsize, self.num_dims))
        eps_s = jax.random.uniform(
            rng_eps_s, (self.popsize,), minval=-1, maxval=1
        )
        idx = jax.random.choice(
            rng_idx, jnp.arange(self.elite_popsize), (self.popsize - 1,)
        )
        x = jnp.concatenate([state.archive[0][None, :], state.archive[idx]])
        sigma_0 = jnp.array(
            [jnp.maximum(params.sigma_best_limit, state.sigma[0])]
        )
        sigma = jnp.concatenate([sigma_0, state.sigma[idx]])
        sigma_gen = sigma * params.sigma_meta ** eps_s
        x += sigma_gen[:, None] * eps_x
        return x, state.replace(archive=x, sigma=sigma_gen)

    def tell_strategy(
        self,
        x: chex.Array,
        fitness: chex.Array,
        state: EvoState,
        params: EvoParams,
    ) -> EvoState:
        """`tell` update to ES state."""
        idx = jnp.argsort(fitness)[: self.elite_popsize]
        fitness = fitness[idx]
        archive = x[idx]
        sigma = state.sigma[idx]

        # Set mean to best member seen so far
        improved = fitness[0] < state.best_fitness
        best_mean = jax.lax.select(improved, archive[0], state.best_member)
        return state.replace(
            fitness=fitness, archive=archive, sigma=sigma, mean=best_mean
        )



from typing import Tuple, Optional, Union
import jax
import jax.numpy as jnp
import chex
import math
from flax import struct
from ..strategy import Strategy


@struct.dataclass
class EvoState:
    mean: chex.Array
    sigma: float
    v: chex.Array
    D: chex.Array
    p_sigma: chex.Array
    p_c: chex.Array
    w_rank_hat: chex.Array
    w_rank: chex.Array
    z: chex.Array
    y: chex.Array
    best_member: chex.Array
    best_fitness: float = jnp.finfo(jnp.float32).max
    gen_counter: int = 0


@struct.dataclass
class EvoParams:
    mu_eff: float
    c_s: float
    c_c: float
    c1: float
    chi_N: float
    h_inv: float
    alpha_dist: float
    lrate_mean: float = 1.0
    lrate_move_sigma: float = 0.1
    lrate_stag_sigma: float = 0.1
    lrate_conv_sigma: float = 0.1
    lrate_B: float = 0.1
    sigma_init: float = 1.0
    init_min: float = 0.0
    init_max: float = 0.0
    clip_min: float = -jnp.finfo(jnp.float32).max
    clip_max: float = jnp.finfo(jnp.float32).max


def get_recombination_weights(popsize: int) -> Tuple[chex.Array, chex.Array]:
    """Get recombination weights for different ranks."""

    def get_weight(i):
        return jnp.log(popsize / 2 + 1) - jnp.log(i)

    w_rank_hat = jax.vmap(get_weight)(jnp.arange(1, popsize + 1))
    w_rank_hat = w_rank_hat * (w_rank_hat >= 0)
    w_rank = w_rank_hat / sum(w_rank_hat) - (1.0 / popsize)
    return w_rank_hat.reshape(-1, 1), w_rank.reshape(-1, 1)


def get_h_inv(dim: int) -> float:
    dim = min(dim, 2000)
    f = lambda a: ((1.0 + a * a) * math.exp(a * a / 2.0) / 0.24) - 10.0 - dim
    f_prime = lambda a: (1.0 / 0.24) * a * math.exp(a * a / 2.0) * (3.0 + a * a)
    h_inv = 1.0
    counter = 0
    while abs(f(h_inv)) > 1e-10:
        counter += 1
        h_inv = h_inv - 0.5 * (f(h_inv) / f_prime(h_inv))
    return h_inv


def w_dist_hat(alpha_dist: float, z: chex.Array) -> chex.Array:
    return jnp.exp(alpha_dist * jnp.linalg.norm(z))


class CR_FM_NES(Strategy):
    def __init__(
        self,
        popsize: int,
        num_dims: Optional[int] = None,
        pholder_params: Optional[Union[chex.ArrayTree, chex.Array]] = None,
        sigma_init: float = 1.0,
        mean_decay: float = 0.0,
        n_devices: Optional[int] = None,
        **fitness_kwargs: Union[bool, int, float]
    ):
        """Cost-Reduced Fast-Moving Natural ES (Nomura & Ono, 2022)
        Reference: https://arxiv.org/abs/2201.11422
        """
        super().__init__(
            popsize,
            num_dims,
            pholder_params,
            mean_decay,
            n_devices,
            **fitness_kwargs
        )
        assert not self.popsize & 1, "Population size must be even"
        self.strategy_name = "CR_FM_NES"

        # Set core kwargs es_params (sigma)
        self.sigma_init = sigma_init

    @property
    def default_params(self) -> EvoParams:
        """Return default parameters of evolutionary strategy."""
        w_rank_hat, w_rank = get_recombination_weights(self.popsize)
        mueff = 1 / (
            (w_rank + (1 / self.popsize)).T @ (w_rank + (1 / self.popsize))
        )
        c_s = (mueff + 2.0) / (self.num_dims + mueff + 5.0)
        c_c = (4.0 + mueff / self.num_dims) / (
            self.num_dims + 4.0 + 2.0 * mueff / self.num_dims
        )
        c1_cma = 2.0 / (jnp.power(self.num_dims + 1.3, 2) + mueff)
        chi_N = jnp.sqrt(self.num_dims) * (
            1.0
            - 1.0 / (4.0 * self.num_dims)
            + 1.0 / (21.0 * self.num_dims * self.num_dims)
        )
        h_inv = get_h_inv(self.num_dims)
        alpha_dist = h_inv * jnp.minimum(
            1.0, jnp.sqrt(self.popsize / self.num_dims)
        )
        lrate_move_sigma = 1.0
        lrate_stag_sigma = jnp.tanh(
            (0.024 * self.popsize + 0.7 * self.num_dims + 20.0)
            / (self.num_dims + 12.0)
        )
        lrate_conv_sigma = 2.0 * jnp.tanh(
            (0.025 * self.popsize + 0.75 * self.num_dims + 10.0)
            / (self.num_dims + 4.0)
        )
        c1 = c1_cma * (self.num_dims - 5) / 6
        lrate_B = jnp.tanh(
            (jnp.minimum(0.02 * self.popsize, 3 * jnp.log(self.num_dims)) + 5)
            / (0.23 * self.num_dims + 25)
        )
        params = EvoParams(
            lrate_move_sigma=lrate_move_sigma,
            lrate_stag_sigma=lrate_stag_sigma,
            lrate_conv_sigma=lrate_conv_sigma,
            lrate_B=lrate_B,
            mu_eff=mueff,
            c_s=c_s,
            c_c=c_c,
            c1=c1,
            chi_N=chi_N,
            alpha_dist=alpha_dist,
            h_inv=h_inv,
            sigma_init=self.sigma_init,
        )
        return params

    def initialize_strategy(
        self, rng: chex.PRNGKey, params: EvoParams
    ) -> EvoState:
        """`initialize` the evolutionary strategy."""
        rng_init, rng_v = jax.random.split(rng)
        initialization = jax.random.uniform(
            rng_init,
            (self.num_dims,),
            minval=params.init_min,
            maxval=params.init_max,
        )
        w_rank_hat, w_rank = get_recombination_weights(self.popsize)
        state = EvoState(
            mean=initialization,
            sigma=params.sigma_init,
            v=jax.random.normal(rng_v, shape=(self.num_dims, 1))
            / jnp.sqrt(self.num_dims),
            D=jnp.ones([self.num_dims, 1]),
            p_sigma=jnp.zeros((self.num_dims, 1)),
            p_c=jnp.zeros((self.num_dims, 1)),
            z=jnp.zeros((self.num_dims, self.popsize)),
            y=jnp.zeros((self.num_dims, self.popsize)),
            w_rank_hat=w_rank_hat.reshape(-1, 1),
            w_rank=w_rank,
            best_member=initialization,
        )

        return state

    def ask_strategy(
        self, rng: chex.PRNGKey, state: EvoState, params: EvoParams
    ) -> Tuple[chex.Array, EvoState]:
        """`ask` for new parameter candidates to evaluate next."""
        z_plus = jax.random.normal(
            rng,
            (int(self.popsize / 2), self.num_dims),
        )
        z = jnp.concatenate([z_plus, -1.0 * z_plus])
        z = jnp.swapaxes(z, 0, 1)
        normv = jnp.linalg.norm(state.v)
        normv2 = normv ** 2
        vbar = state.v / normv

        # Rescale/reparametrize noise
        y = z + (jnp.sqrt(1 + normv2) - 1) * vbar @ (vbar.T @ z)
        x = state.mean[:, None] + state.sigma * y * state.D
        x = jnp.swapaxes(x, 0, 1)
        return x, state.replace(z=z, y=y)

    def tell_strategy(
        self,
        x: chex.Array,
        fitness: chex.Array,
        state: EvoState,
        params: EvoParams,
    ) -> EvoState:
        """`tell` performance data for strategy state update."""
        ranks = fitness.argsort()
        z = state.z[:, ranks]
        y = state.y[:, ranks]
        x = jnp.swapaxes(x, 0, 1)[:, ranks]

        # Update evolution path p_sigma
        p_sigma = (1 - params.c_s) * state.p_sigma + jnp.sqrt(
            params.c_s * (2.0 - params.c_s) * params.mu_eff
        ) * (z @ state.w_rank)
        p_sigma_norm = jnp.linalg.norm(p_sigma)

        # Calculate distance weight
        w_tmp = state.w_rank_hat * jax.vmap(w_dist_hat, in_axes=(None, 1))(
            params.alpha_dist, z
        ).reshape(-1, 1)
        weights_dist = w_tmp / sum(w_tmp) - 1.0 / self.popsize

        # switching weights and learning rate
        p_sigma_cond = p_sigma_norm >= params.chi_N
        weights = jax.lax.select(p_sigma_cond, weights_dist, state.w_rank)
        lrate_sigma = jax.lax.select(
            p_sigma_cond, params.lrate_move_sigma, params.lrate_stag_sigma
        )
        lrate_sigma = jax.lax.select(
            p_sigma_norm >= 0.1 * params.chi_N,
            lrate_sigma,
            params.lrate_conv_sigma,
        )

        # update evolution path p_c and mean
        wxm = (x - state.mean[:, None]) @ weights
        p_c = (1.0 - params.c_c) * state.p_c + jnp.sqrt(
            params.c_c * (2.0 - params.c_c) * params.mu_eff
        ) * wxm / state.sigma
        mean = state.mean + params.lrate_mean * wxm.squeeze()

        normv = jnp.linalg.norm(state.v)
        vbar = state.v / normv
        normv2 = normv ** 2
        normv4 = normv2 ** 2

        exY = jnp.append(y, p_c / state.D, axis=1)
        yy = exY * exY
        ip_yvbar = vbar.T @ exY
        yvbar = exY * vbar
        gammav = 1.0 + normv2
        vbarbar = vbar * vbar
        alphavd = jnp.minimum(
            1,
            jnp.sqrt(
                normv4 + (2 * gammav - jnp.sqrt(gammav)) / jnp.max(vbarbar)
            )
            / (2 + normv2),
        )
        t = exY * ip_yvbar - vbar * (ip_yvbar ** 2 + gammav) / 2
        b = -(1 - alphavd ** 2) * normv4 / gammav + 2 * alphavd ** 2
        H = jnp.ones([self.num_dims, 1]) * 2 - (b + 2 * alphavd ** 2) * vbarbar
        invH = H ** (-1)
        s_step1 = (
            yy
            - normv2 / gammav * (yvbar * ip_yvbar)
            - jnp.ones([self.num_dims, self.popsize + 1])
        )
        ip_vbart = vbar.T @ t
        s_step2 = s_step1 - alphavd / gammav * (
            (2 + normv2) * (t * vbar) - normv2 * vbarbar @ ip_vbart
        )
        invHvbarbar = invH * vbarbar
        ip_s_step2invHvbarbar = invHvbarbar.T @ s_step2
        s = (s_step2 * invH) - b / (
            1 + b * vbarbar.T @ invHvbarbar
        ) * invHvbarbar @ ip_s_step2invHvbarbar
        ip_svbarbar = vbarbar.T @ s
        t = t - alphavd * ((2 + normv2) * (s * vbar) - vbar @ ip_svbarbar)

        # update v, D covariance ingredients
        exw = jnp.append(
            params.lrate_B * weights,
            jnp.array([params.c1]).reshape(1, 1),
            axis=0,
        )
        v = state.v + (t @ exw) / normv
        D = state.D + (s @ exw) * state.D
        # calculate detA
        nthrootdetA = jnp.exp(
            jnp.sum(jnp.log(D)) / self.num_dims
            + jnp.log(1 + v.T @ v) / (2 * self.num_dims)
        )[0][0]
        D = D / nthrootdetA
        # update sigma
        G_s = (
            jnp.sum((z * z - jnp.ones([self.num_dims, self.popsize])) @ weights)
            / self.num_dims
        )
        sigma = state.sigma * jnp.exp(lrate_sigma / 2 * G_s)
        return state.replace(
            p_sigma=p_sigma, mean=mean, p_c=p_c, v=v, D=D, sigma=sigma
        )



import jax
import jax.numpy as jnp
import chex
from typing import Tuple, Optional
from functools import partial


class SequenceFitness(object):
    def __init__(
        self,
        task_name: str = "SeqMNIST",
        batch_size: int = 128,
        seq_length: int = 150,  # Sequence length in addition task
        permute_seq: bool = False,  # Permuted S-MNIST task option
        test: bool = False,
        n_devices: Optional[int] = None,
    ):
        self.task_name = task_name
        self.batch_size = batch_size
        self.steps_per_member = 1
        self.test = test

        # Setup task-specific input/output shapes and loss fn
        if self.task_name == "SeqMNIST":
            self.action_shape = 10
            self.permute_seq = permute_seq
            self.seq_length = 784
            self.loss_fn = partial(loss_and_acc, num_classes=10)
        elif self.task_name == "Addition":
            self.action_shape = 1
            self.permute_seq = False
            self.seq_length = seq_length
            self.loss_fn = loss_and_mae
        else:
            raise ValueError("Dataset is not supported.")

        data = get_array_data(
            self.task_name, self.seq_length, self.permute_seq, self.test
        )
        self.dataloader = BatchLoader(*data, batch_size=self.batch_size)
        self.num_rnn_steps = self.dataloader.data_shape[1]

        if n_devices is None:
            self.n_devices = jax.local_device_count()
        else:
            self.n_devices = n_devices

    def set_apply_fn(self, network, carry_init):
        """Set the network forward function."""
        self.network = network
        self.carry_init = carry_init
        self.rollout_pop = jax.vmap(self.rollout_rnn, in_axes=(None, 0))
        # pmap over popmembers if > 1 device is available - otherwise pmap
        if self.n_devices > 1:
            self.rollout = self.rollout_pmap
            print(
                f"SequenceFitness: {self.n_devices} devices detected. Please"
                " make sure that the ES population size divides evenly across"
                " the number of devices to pmap/parallelize over."
            )
        else:
            self.rollout = jax.jit(self.rollout_vmap)

    def rollout_vmap(
        self, rng_input: chex.PRNGKey, network_params: chex.ArrayTree
    ):
        """Vectorize rollout. Reshape output correctly."""
        loss, perf = self.rollout_pop(rng_input, network_params)
        loss_re = loss.reshape(-1, 1)
        perf_re = perf.reshape(-1, 1)
        return loss_re, perf_re

    def rollout_pmap(
        self, rng_input: chex.PRNGKey, network_params: chex.ArrayTree
    ):
        """Parallelize rollout across devices. Split keys/reshape correctly."""
        keys_pmap = jnp.tile(rng_input, (self.n_devices, 1))
        loss_dev, perf_dev = jax.pmap(self.rollout_pop)(
            keys_pmap, network_params
        )
        loss_re = loss_dev.reshape(-1, 1)
        perf_re = perf_dev.reshape(-1, 1)
        return loss_re, perf_re

    def rollout_rnn(
        self, rng_input: chex.PRNGKey, network_params: chex.ArrayTree
    ) -> Tuple[float, float]:
        """Evaluate a network on a supervised learning task."""
        rng, rng_sample = jax.random.split(rng_input)
        X, y = self.dataloader.sample(rng_sample)
        # Map over sequence batch dimension
        y_pred = jax.vmap(self.rollout_single, in_axes=(None, None, 0))(
            rng, network_params, X
        )
        loss, perf = self.loss_fn(y_pred, y)
        # Return negative loss to maximize!
        return -1 * loss, perf

    def rollout_single(
        self,
        rng: chex.PRNGKey,
        network_params: chex.ArrayTree,
        X_single: chex.ArrayTree,
    ):
        """Rollout RNN on a single sequence."""
        # Reset the network
        hidden = self.carry_init()

        def rnn_step(state_input, tmp):
            """lax.scan compatible step transition in jax env."""
            network_params, hidden, rng, t = state_input
            rng, rng_net = jax.random.split(rng)
            hidden, pred = self.network(
                network_params,
                X_single[t],
                hidden,
                rng_net,
            )
            carry = [network_params, hidden, rng, t + 1]
            return carry, pred

        # Scan over image length (784)/sequence
        _, scan_out = jax.lax.scan(
            rnn_step, [network_params, hidden, rng, 0], (), self.num_rnn_steps
        )
        y_pred = scan_out[-1]
        return y_pred

    @property
    def input_shape(self) -> Tuple[int]:
        """Get the shape of the observation."""
        return self.dataloader.data_shape


def loss_and_acc(
    y_pred: chex.Array, y_true: chex.Array, num_classes: int
) -> Tuple[chex.Array, chex.Array]:
    """Compute cross-entropy loss and accuracy."""
    acc = jnp.mean(jnp.argmax(y_pred, axis=-1) == y_true)
    labels = jax.nn.one_hot(y_true, num_classes)
    loss = -jnp.sum(labels * jax.nn.log_softmax(y_pred))
    loss /= labels.shape[0]
    return loss, acc


def loss_and_mae(
    y_pred: chex.Array, y_true: chex.Array
) -> Tuple[chex.Array, chex.Array]:
    """Compute mean squared error loss and mean absolute error."""
    loss = jnp.mean((y_pred.squeeze() - y_true) ** 2)
    mae = jnp.mean(jnp.abs(y_pred.squeeze() - y_true))
    return loss, -mae


class BatchLoader:
    def __init__(
        self,
        X: chex.Array,
        y: chex.Array,
        batch_size: int,
    ):
        self.X = X
        self.y = y
        self.data_shape = self.X.shape[1:][::-1]
        self.num_train_samples = X.shape[0]
        self.batch_size = batch_size

    def sample(self, key: chex.PRNGKey) -> Tuple[chex.Array, chex.Array]:
        """Sample a single batch of X, y data."""
        sample_idx = jax.random.choice(
            key,
            jnp.arange(self.num_train_samples),
            (self.batch_size,),
            replace=False,
        )
        return (
            jnp.take(self.X, sample_idx, axis=0),
            jnp.take(self.y, sample_idx, axis=0),
        )


def get_smnist_loaders(test: bool = False):
    try:
        import torch
        from torchvision import datasets, transforms
    except ModuleNotFoundError as err:
        raise ModuleNotFoundError(
            f"{err}. You need to install `torch` and `torchvision`"
            "to use the `SupervisedFitness` module."
        )

    transform = transforms.Compose(
        [
            transforms.ToTensor(),
            transforms.Normalize((0.1307,), (0.3081,)),
            transforms.Lambda(lambda x: torch.flatten(x)),
            transforms.Lambda(lambda x: torch.unsqueeze(x, -1)),
        ]
    )
    bs = 10000 if test else 60000
    loader = torch.utils.data.DataLoader(
        datasets.MNIST(
            "~/data", download=True, train=not test, transform=transform
        ),
        batch_size=bs,
        shuffle=False,
    )
    return loader


def get_adding_data(T: int = 150, test: bool = False):
    """
    Sample a mask, [0, 1] samples and sum of targets for len T.
    Reference:  Martens & Sutskever. ICML, 2011.
    """
    rng = jax.random.PRNGKey(0)
    bs = 100000 if test else 10000

    def get_single_addition(rng, T):
        rng_numb, rng_mask = jax.random.split(rng)
        numbers = jax.random.uniform(rng_numb, (T,), minval=0, maxval=1)
        mask_ids = jax.random.choice(
            rng_mask, jnp.arange(T), (2,), replace=False
        )
        mask = jnp.zeros(T).at[mask_ids].set(1)
        target = jnp.sum(mask * numbers)
        return jnp.stack([numbers, mask], axis=1), target

    batch_seq_gen = jax.vmap(get_single_addition, in_axes=(0, None))
    data, target = batch_seq_gen(jax.random.split(rng, bs), T)
    return data, target


def get_array_data(
    task_name: str = "SMNIST",
    seq_length: int = 150,
    permute_seq: bool = False,
    test: bool = False,
):
    """Get raw data arrays to subsample from."""
    if task_name == "SeqMNIST":
        loader = get_smnist_loaders(test)
        for _, (data, target) in enumerate(loader):
            break
        data, target = jnp.array(data), jnp.array(target)

        # Permute the sequence of the pixels if desired.
        if permute_seq:  # bs, T - fix permutation by seed
            rng = jax.random.PRNGKey(0)
            idx = jnp.arange(784)
            idx_perm = jax.random.permutation(rng, idx)
            data = data.at[:].set(data[:, idx_perm])
    elif task_name == "Addition":
        data, target = get_adding_data(seq_length, test)
        data, target = jnp.array(data), jnp.array(target)
    else:
        raise ValueError("Dataset is not supported.")
    return data, target



from typing import Tuple, Optional, Union
import jax
import jax.numpy as jnp
import chex
from flax import struct
from ..strategy import Strategy


@struct.dataclass
class EvoState:
    mean: chex.Array
    archive: chex.Array
    fitness_archive: chex.Array
    best_member: chex.Array
    best_fitness: float = jnp.finfo(jnp.float32).max
    gen_counter: int = 0


@struct.dataclass
class EvoParams:
    mutate_best_vector: bool = True  # False - 'random'
    num_diff_vectors: int = 1  # [1, 2]
    cross_over_rate: float = 0.9  # cross-over probability [0, 1]
    diff_w: float = 0.8  # differential weight (F) [0, 2]
    init_min: float = -0.1
    init_max: float = 0.1
    clip_min: float = -jnp.finfo(jnp.float32).max
    clip_max: float = jnp.finfo(jnp.float32).max


class DE(Strategy):
    def __init__(
        self,
        popsize: int,
        num_dims: Optional[int] = None,
        pholder_params: Optional[Union[chex.ArrayTree, chex.Array]] = None,
        n_devices: Optional[int] = None,
        **fitness_kwargs: Union[bool, int, float]
    ):
        """Differential Evolution (Storn & Price, 1997)
        Reference: https://tinyurl.com/4pje5a74"""
        assert popsize > 6, "DE requires popsize > 6."
        super().__init__(
            popsize,
            num_dims,
            pholder_params,
            n_devices=n_devices,
            **fitness_kwargs
        )
        self.strategy_name = "DE"

    @property
    def params_strategy(self) -> EvoParams:
        return EvoParams()

    def initialize_strategy(
        self, rng: chex.PRNGKey, params: EvoParams
    ) -> EvoState:
        """
        `initialize` the differential evolution strategy.
        Initialize all population members by randomly sampling
        positions in search-space (defined in `params`).
        """
        initialization = jax.random.uniform(
            rng,
            (self.popsize, self.num_dims),
            minval=params.init_min,
            maxval=params.init_max,
        )
        state = EvoState(
            mean=initialization.mean(axis=0),
            archive=initialization,
            fitness_archive=jnp.zeros(self.popsize) + 20e10,
            best_member=initialization.mean(axis=0),
        )
        return state

    def ask_strategy(
        self, rng: chex.PRNGKey, state: EvoState, params: EvoParams
    ) -> Tuple[chex.Array, EvoState]:
        """
        `ask` for new proposed candidates to evaluate next.
        For each population member x:
        - Pick 3 unique members (a, b, c) from rest of pop. at random.
        - Pick a random dimension R from the parameter space.
        - Compute the member's potentially new position:
            - For each i ∈ {1,...,d}, pick a r_i ~ U(0, 1)
            - If r_i < cr or i = R, set y_i = a_i + F * (b_i - c_i)
            - Else y_i = x_i
        Return new potential position y.
        """
        rng_members = jax.random.split(rng, self.popsize)
        member_ids = jnp.arange(self.popsize)
        x = jax.vmap(single_member_ask, in_axes=(0, 0, None, None, None, None))(
            rng_members,
            member_ids,
            self.num_dims,
            state.archive,
            state.best_member,
            params,
        )
        return jnp.squeeze(x, axis=2), state

    def tell_strategy(
        self,
        x: chex.Array,
        fitness: chex.Array,
        state: chex.ArrayTree,
        params: EvoParams,
    ) -> EvoState:
        """
        `tell` update to ES state.
        If fitness of y <= fitness of x -> replace in population.
        """
        # Replace member in archive if performance was improved
        replace = fitness <= state.fitness_archive
        archive = (
            jnp.expand_dims(replace, 1) * x
            + (1 - jnp.expand_dims(replace, 1)) * state.archive
        )
        fitness_archive = (
            replace * fitness + (1 - replace) * state.fitness_archive
        )
        # Keep mean across stored archive around for evaluation protocol
        mean = archive.mean(axis=0)
        return state.replace(
            mean=mean, archive=archive, fitness_archive=fitness_archive
        )


def single_member_ask(
    rng: chex.PRNGKey,
    member_id: int,
    num_dims: int,
    archive: chex.Array,
    best_member: chex.Array,
    params: EvoParams,
) -> chex.Array:
    """Perform `ask` steps for single member."""
    x = archive[member_id]

    # Sample a, b and c parameter vectors from rest of population
    rng, rng_vectors, rng_R = jax.random.split(rng, 3)
    # A bit of an awkward hack - sample one additional member to avoid
    # using same vector as x - check condition and select extra if needed
    # Also always sample 6 members - for case where we want two diff vectors
    row_ids = jax.random.choice(
        rng_vectors, jnp.arange(archive.shape[0]), (6,), replace=False
    )
    a = jax.lax.select(
        row_ids[0] == member_id, archive[row_ids[5]], archive[row_ids[0]]
    )
    b = jax.lax.select(
        row_ids[1] == member_id, archive[row_ids[5]], archive[row_ids[1]]
    )
    c = jax.lax.select(
        row_ids[2] == member_id, archive[row_ids[5]], archive[row_ids[2]]
    )
    d = jax.lax.select(
        row_ids[3] == member_id, archive[row_ids[5]], archive[row_ids[3]]
    )
    e = jax.lax.select(
        row_ids[4] == member_id, archive[row_ids[5]], archive[row_ids[4]]
    )

    # Use best vector instead of random `a` vector if `mutate_vector` == "best"
    a = jax.lax.select(params.mutate_best_vector, best_member, a)

    # Sample random dimension that will be alter for sure
    R = jax.random.randint(rng_R, (1,), minval=0, maxval=num_dims)

    rng_dims = jax.random.split(rng, num_dims)
    dim_ids = jnp.arange(num_dims)
    y = jax.vmap(
        single_dimension_ask,
        in_axes=(
            0,
            0,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
        ),
    )(
        rng_dims,
        dim_ids,
        x,
        a,
        b,
        c,
        d,
        e,
        R,
        params.cross_over_rate,
        params.diff_w,
        params.num_diff_vectors == 2,
    )
    return y


def single_dimension_ask(
    rng: chex.PRNGKey,
    dim_id: int,
    x: chex.Array,
    a: chex.Array,
    b: chex.Array,
    c: chex.Array,
    d: chex.Array,
    e: chex.Array,
    R: int,
    cr: float,
    diff_w: float,
    use_second_diff: bool,
) -> chex.Array:
    """Perform `ask` step for single dimension."""
    r_i = jax.random.uniform(rng, (1,))
    mutate_bool = jnp.logical_or(r_i < cr, dim_id == R)
    y_i = (
        mutate_bool * a[dim_id]
        + diff_w * (b[dim_id] - c[dim_id])
        # Only add second difference vector if desired!
        + diff_w * (d[dim_id] - e[dim_id]) * use_second_diff
        # Mutation - exchange x dim with a dim.
        + (1 - mutate_bool) * x[dim_id]
    )
    return y_i
